"""CGCNN Wrapper."""

from __future__ import annotations

import argparse
import os
import warnings
from typing import TYPE_CHECKING

import numpy as np
from monty.dev import requires

from maml.apps.bowsr.model.base import EnergyModel

if TYPE_CHECKING:
    from pymatgen.core.structure import Structure

try:
    import cgcnn
    import torch
    from cgcnn.data import AtomCustomJSONInitializer, GaussianDistance
    from cgcnn.model import CrystalGraphConvNet
    from torch import Tensor
except ImportError:
    torch = None  # type: ignore
    cgcnn = None  # type: ignore
    Tensor = None  # type: ignore

pjoin = os.path.join
module_dir = os.path.dirname(__file__)
model_dir = pjoin(module_dir, "model_files", "cgcnn", "formation-energy-per-atom.pth.tar")


@requires(cgcnn is not None and torch is not None, "cgcnn and torch are needed to use the CGCNN evaluator.")
class CGCNN(EnergyModel):
    """Wrapper to generate cgcnn energy prediction model."""

    def __init__(self, model_path: str = model_dir, orig_atom_fea_len: int = 92, nbr_fea_len: int = 41):
        """
        Init CGCNN.

        Args:
            model_path(str): path of model
            orig_atom_fea_len(int): Number of atom features in the input.
                                    i.e. Original atom feature length
                                    (default 92)
            nbr_fea_len(int): Number of bond features.
                            i.e. Number of neighbors (default 41).

        """
        checkpoint = torch.load(model_path, map_location=lambda storage, loc: storage)
        model_args = argparse.Namespace(**checkpoint["args"])
        self.model = CrystalGraphConvNet(
            orig_atom_fea_len=orig_atom_fea_len,
            nbr_fea_len=nbr_fea_len,
            n_h=model_args.n_h,
            n_conv=model_args.n_conv,
            h_fea_len=model_args.h_fea_len,
        )
        self.normalizer = CGCNNNormalizer(torch.zeros(3))
        self.model.load_state_dict(checkpoint["state_dict"])
        self.normalizer.load_state_dict(checkpoint["normalizer"])

    def predict_energy(self, structure: Structure) -> np.ndarray:  # type: ignore
        """
        CGCNN predict formatio nenergy from pymatgen structure.

        Args:
            structure(Structure): structure to be predicted

        Returns: formation energy (eV/atom) of provided structure

        """
        self.model.eval()
        input_generator = CGCNNInput()
        inp = input_generator.generate_input(structure)
        inp = (*inp, [torch.LongTensor(np.arange(structure.num_sites))])  # type: ignore
        output = self.model(*inp)
        return self.normalizer.denorm(output).data.cpu().numpy()[0][0]


@requires(cgcnn is not None and torch is not None, "cgcnn and torch are needed to use the CGCNN evaluator.")
class CGCNNInput:
    """Wrapper to generate input for cgcnn from pymatgen structure."""

    atom_init_filename = pjoin(module_dir, "model_files", "cgcnn", "atom_init.json")

    def __init__(self, max_num_nbr: int = 12, radius: float = 8, dmin: float = 0, step: float = 0.2):
        """
        Init CGCNNInput.

        Args:
            max_num_nbr(int): The maximum number of neighbors
                            while constructing the crystal graph
                            (default 12)
            radius(float): The cutoff radius for searching neighbors
                            (default 8)
            dmin(float): The minimum distance for constructing
                        GaussianDistance (default 0)
            step(float): The step size for constructing GaussianDistance
                        (default 0.2)
        """
        self.max_num_nbr = max_num_nbr
        self.radius = radius
        self.dmin = dmin
        self.step = step
        self.gdf = GaussianDistance(dmin=dmin, dmax=self.radius, step=step)
        self.ari = AtomCustomJSONInitializer(self.atom_init_filename)

    def _get_nbr_fea(self, all_nbrs: list, cif_id: int) -> tuple[np.ndarray, ...]:
        nbr_fea_idx, nbr_fea = [], []
        for nbr in all_nbrs:
            if len(nbr) < self.max_num_nbr:
                warnings.warn(
                    f"{cif_id} not find enough neighbors to build graph. "
                    "If it happens frequently, consider increase "
                    "radius."
                )
                nbr_fea_idx.append([x[2] for x in nbr] + [0] * (self.max_num_nbr - len(nbr)))
                nbr_fea.append([x[1] for x in nbr] + [self.radius + 1.0] * (self.max_num_nbr - len(nbr)))
            else:
                nbr_fea_idx.append([x[2] for x in nbr[: self.max_num_nbr]])
                nbr_fea.append([x[1] for x in nbr[: self.max_num_nbr]])
        nbr_fea_idx, nbr_fea = np.array(nbr_fea_idx), np.array(nbr_fea)  # type: ignore
        nbr_fea = self.gdf.expand(nbr_fea)
        return (nbr_fea_idx, nbr_fea)  # type: ignore

    def generate_input(self, structure: Structure, cif_id: int | None = None) -> tuple:
        """
        Generate cgcnn inputs for given structure.

        Args:
            structure(Structure): structure to get input for
            cif_id(int): Optional, the id of the structure

        Returns: Tuple of input (atom_fea, nbr_fea, nbr_fea_idx)
        """
        atom_fea = [sum((self.ari.get_atom_fea(el.Z) * oc) for el, oc in site.species.items()) for site in structure]
        atom_fea = np.vstack(atom_fea)  # type: ignore
        atom_fea = Tensor(atom_fea)  # type: ignore
        all_nbrs = structure.get_all_neighbors(self.radius, include_index=True)
        # sort the nbrs by distance
        all_nbrs = [sorted(nbrs, key=lambda x: x[1]) for nbrs in all_nbrs]

        if cif_id:
            nbr_fea_idx, nbr_fea = self._get_nbr_fea(all_nbrs, cif_id)
        else:
            nbr_fea_idx, nbr_fea = self._get_nbr_fea(all_nbrs, 0)
        nbr_fea = Tensor(nbr_fea)  # type: ignore
        nbr_fea_idx = torch.LongTensor(nbr_fea_idx)  # type: ignore
        return (atom_fea, nbr_fea, nbr_fea_idx)

    def generate_inputs(self, structures: list[Structure], cif_ids: list[int] | None = None) -> list[tuple]:
        """
        Generate cgcnn inputs for given list of structures
        Args:
            structures (list): List of structures to get inputs for.
            cif_ids (list): Optional, the list of ids of the structures.

        """
        if not cif_ids:
            cif_ids = list(range(len(structures)))
        return [self.generate_input(s, id) for s, id in zip(structures, cif_ids, strict=False)]


@requires(cgcnn is not None and torch is not None, "cgcnn and torch are needed to use the CGCNN evaluator.")
class CGCNNNormalizer:
    """Normalize a Tensor and restore it later."""

    def __init__(self, tensor: Tensor):
        """
        Tensor is taken as a sample to calculate the mean and std.

        Args:
            tensor(Tensor): data
        """
        self.mean = torch.mean(tensor)
        self.std = torch.std(tensor)

    def norm(self, tensor: Tensor) -> Tensor:
        """
        Normalize tensor.

        Args:
            tensor(Tensor): data

        Returns: normalized tensor

        """
        return (tensor - self.mean) / self.std

    def denorm(self, normed_tensor: Tensor) -> Tensor:
        """
        Denormalize tensor.

        Args:
            normed_tensor(Tensor): normalized tensor data

        Returns: denormalized tensor

        """
        return normed_tensor * self.std + self.mean

    def state_dict(self) -> dict:
        """
        Get dict of mean and std.

        Returns: dict of mean and std of the normalizer。

        """
        return {"mean": self.mean, "std": self.std}

    def load_state_dict(self, state_dict: dict) -> None:
        """
        Load the normalizer with mean and std.

        Args:
            state_dict(Dict): dict of mean and std

        Returns: None

        """
        self.mean = state_dict["mean"]
        self.std = state_dict["std"]



"""MAML data base classes."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pandas as pd


class BaseDataSource:
    """Abstract base class for a data source."""

    @abc.abstractmethod
    def get(self, *args, **kwargs) -> pd.DataFrame:
        """Get data from sources."""



"""Clustering methods."""

from __future__ import annotations

import logging

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.cluster import Birch

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BirchClustering(BaseEstimator, TransformerMixin):
    """Birch Clustering as one step of the DIRECT pipeline."""

    def __init__(self, n=None, threshold_init=0.5, **kwargs):
        """
        Args:
            n: Clustering the PCs into n clusters. When n is None, the number of clusters
                is dependent on threshold_init and other kwargs, and the final
                (global) clustering step is skipped. Default to None.
            threshold_init: The initial radius of the subcluster obtained by merging
                a new sample and the closest subcluster should be lesser than
                the threshold. Otherwise, a new subcluster is started. See details in:
                https://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html.
                Users may tune this value for desired performance of birch, while 0.5
                is generally a good starting point, and some automatic tuning is done
                with our built-in codes to achieve n clusters if given.
            **kwargs: Pass to BIRCH.
        """
        self.n = n
        self.threshold_init = threshold_init
        self.kwargs = kwargs

    def fit(self, X, y=None):
        """
        Place holder for fit API.

        Args:
            X: Any inputs
            y: Any outputs

        Returns: self
        """
        return self

    def transform(self, PCAfeatures):
        """
        Perform Birch Clustering to an array of input PCA features.

        Args:
            PCAfeatures: An array of PCA features.

        Returns:
            A dict of Birch Clustering results, including labels of each
            PCA feature, centroid positions of each cluster in PCA feature s
            pace, and the array of input PCA features.
        """
        model = Birch(n_clusters=self.n, threshold=self.threshold_init, **self.kwargs).fit(PCAfeatures)
        if self.n is not None:
            while (
                len(set(model.subcluster_labels_)) < self.n
            ):  # decrease threshold until desired n clusters is achieved
                logger.info(
                    f"BirchClustering with threshold_init={self.threshold_init} and n={self.n} "
                    f"gives {len(set(model.subcluster_labels_))} clusters.",
                )
                self.threshold_init = self.threshold_init / self.n * len(set(model.subcluster_labels_))
                model = Birch(n_clusters=self.n, threshold=self.threshold_init, **self.kwargs).fit(PCAfeatures)

        labels = model.predict(PCAfeatures)
        self.model = model
        logger.info(
            f"BirchClustering with threshold_init={self.threshold_init} and n={self.n} "
            f"gives {len(set(model.subcluster_labels_))} clusters.",
        )
        label_centers = dict(zip(model.subcluster_labels_, model.subcluster_centers_, strict=False))
        return {
            "labels": labels,
            "label_centers": label_centers,
            "PCAfeatures": PCAfeatures,
        }



"""Model mixins."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from maml.utils import DummyScaler

if TYPE_CHECKING:
    from ._model import BaseModel


class TargetScalerMixin:
    """Mixin class for models with target scaler."""

    def __init__(self, *args, **kwargs):
        """
        Mixin with target scaler, required kwarg target_scaler.

        Args:
            *args: input args for model constructor
            **kwargs: kwargs for model constructor
        """
        target_scaler = kwargs.pop("target_scaler", DummyScaler())
        self.target_scaler = target_scaler
        super().__init__(*args, **kwargs)

    def train(
        self,
        objs: list | np.ndarray,
        targets: list | np.ndarray | None = None,
        val_objs: list | np.ndarray | None = None,
        val_targets: list | np.ndarray | None = None,
        **kwargs,
    ) -> BaseModel:
        """
        Train the models from object, target pairs.

        Args:
            objs (list of objects): List of objects
            targets (list): list of float or np.ndarray
            val_objs (list of objects): list of validation objects
            val_targets (list): list of validation targets
            **kwargs: kwargs for the train method.

        Returns: self

        """
        if targets is None:
            return super().train(objs, targets, val_objs, val_targets, **kwargs)  # type: ignore

        num_atoms = [len(i) for i in objs]
        targets = [self.target_scaler.transform(i, j) for i, j in zip(targets, num_atoms, strict=False)]
        if val_objs is not None and val_targets is not None:
            num_val_atoms = [len(i) for i in val_objs]
            val_targets = [self.target_scaler.transform(i, j) for i, j in zip(val_targets, num_val_atoms, strict=False)]
        super().train(objs, targets, val_objs, val_targets, **kwargs)  # type: ignore
        return self  # type: ignore

    def predict_objs(self, objs: list | np.ndarray) -> np.ndarray:
        """
        Predict the values given a set of objects. Usually Pymatgen
            Structure objects.
        """
        preds = super().predict_objs(objs)  # type: ignore
        ns = [len(i) for i in objs]
        return np.array([self.target_scaler.inverse_transform(i, j) for i, j in zip(preds, ns, strict=False)])



"""DIRECT sampling."""

from __future__ import annotations

from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from maml.describers import M3GNetStructure

from .clustering import BirchClustering
from .pca import PrincipalComponentAnalysis
from .stratified_sampling import SelectKFromClusters


class DIRECTSampler(Pipeline):
    """
    DImensionality REduction-Clustering-sTratified (DIRECT)
    sampling Pipeline. For more details, please refer to our
    manuscript: https://arxiv.org/abs/2307.13710.
    """

    def __init__(
        self,
        structure_encoder="M3GNet",
        scaler="StandardScaler",
        pca="PrincipalComponentAnalysis",
        weighting_PCs=True,
        clustering="Birch",
        select_k_from_clusters="select_k_from_clusters",
    ):
        """
        Args:
        structure_encoder: Structure featurizer. It can be any encoder
            that takes in a list of N structures and returns a 2-D array
            of N*D features, where D is the fixed dimensionality of the
            feature vector. By default, the M3GNet formation energy model
            is used. Set this to False to skip the encoding step when needed.
        scaler: StandardScaler to perform normalization before PCA.
        pca: PCA for dimensionality reduction.
        weighting_PCs: Whether to weight PC with their explained variance.
        clustering: Clustering method to clustering based on PCs.
        select_k_from_clusters: Straitified sampling of k structures from
            each cluster.
        """
        self.structure_encoder = M3GNetStructure() if structure_encoder == "M3GNet" else structure_encoder
        self.scaler = StandardScaler() if scaler == "StandardScaler" else scaler
        self.pca = (
            PrincipalComponentAnalysis(weighting_PCs=weighting_PCs) if pca == "PrincipalComponentAnalysis" else pca
        )
        self.weighting_PCs = weighting_PCs
        self.clustering = BirchClustering() if clustering == "Birch" else clustering
        self.select_k_from_clusters = (
            SelectKFromClusters() if select_k_from_clusters == "select_k_from_clusters" else select_k_from_clusters
        )
        steps = [
            (i.__class__.__name__, i)
            for i in [
                self.structure_encoder,
                self.scaler,
                self.pca,
                self.clustering,
                self.select_k_from_clusters,
            ]
            if i
        ]
        super().__init__(steps)



"""
Radial distribution functions for site features.
This was originally written in pymatgen-diffusion.
"""

from __future__ import annotations

from math import ceil
from typing import TYPE_CHECKING

import numpy as np
from scipy.ndimage import gaussian_filter1d

if TYPE_CHECKING:
    from pymatgen.core import Structure


class RadialDistributionFunction:
    """Calculator for radial distribution function."""

    def __init__(
        self,
        r_min: float = 0.0,
        r_max: float = 10.0,
        n_grid: int = 101,
        sigma: float = 0.0,
    ):
        """
        Fast radial distribution analysis. This method calculates
        rdf on `np.linspace(r_min, r_max, n_grid)` points.

        Args:
            r_min (float): minimal radius
            r_max (float): maximal radius
            n_grid (int): number of grid points, defaults to 101
            sigma (float): smooth parameter.
        """
        self.r_min = r_min
        self.r_max = r_max
        self.n_grid = n_grid

        self.dr = (self.r_max - self.r_min) / (self.n_grid - 1)  # end points are on grid
        self.r = np.linspace(self.r_min, self.r_max, self.n_grid)

        self.cutoff = self.r_max + self.dr / 2.0  # add a small shell to improve robustness
        self.sigma = ceil(sigma / self.dr)
        self.volumes = 4.0 * np.pi * self.r**2 * self.dr
        self.volumes[self.volumes < 1e-8] = 1e8  # avoid divide by zero

    def get_site_rdf(self, structure: Structure) -> tuple[np.ndarray, list[dict]]:
        """
        Args:
            structure (Structure): pymatgen structure
        Returns:
            r, rdfs, r is the radial points, and rdfs are a list of rdf dicts
            rdfs[0] is the rdfs of the first site. It is a dictionary of {atom_pair: pair_rdf}
            e.g.,
                {"Sr:O": [0, 0, 0.1, 0.2, ..]}.
        """
        pair_distances = get_pair_distances(structure, self.cutoff)
        density = self._get_specie_density(structure)

        # return full rdf information for site

        rdfs: list[dict] = [{}] * len(structure)
        for i, neighbors in enumerate(pair_distances):
            if len(neighbors["neighbors"]) == 0:
                continue
            c_specie = neighbors["specie"]
            temp_neighbors = neighbors["neighbors"]

            rdfs[i] = {
                f"{c_specie}:{specie}": _dist_to_counts(
                    temp_neighbors[specie],
                    r_min=self.r_min,
                    r_max=self.r_max,
                    n_grid=self.n_grid,
                )
                / self.volumes
                / density[specie]
                for specie in temp_neighbors
            }

            if self.sigma > 1e-8:
                rdfs[i] = {key: gaussian_filter1d(rdfs[i][key], self.sigma) for key in rdfs[i]}
        return self.r, rdfs

    def get_species_rdf(
        self,
        structure: Structure,
        ref_species: list | None = None,
        species: list | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Get specie-wise rdf
        Args:
            structure (Structure): target structure
            ref_species (list of species or just single specie str): the reference species.
                The rdfs are calculated with these species at the center
            species (list of species or just single specie str): the species that we are interested in.
                The rdfs are calculated on these species.

        Returns: 2-D array of distances and rdf intensities.
        """
        all_species = list({str(i.specie) for i in structure.sites})
        density = self._get_specie_density(structure)
        n_atoms = structure.composition.to_data_dict["unit_cell_composition"]
        if ref_species is None:
            ref_species = all_species
        if species is None:
            species = all_species

        pair_distances = get_pair_distances(structure, self.cutoff)
        all_distances = []
        for _i, neighbors in enumerate(pair_distances):
            if neighbors["specie"] not in ref_species:
                continue

            for n_specie, distances in neighbors["neighbors"].items():
                if n_specie not in species:
                    continue
                all_distances.append(distances)

        if len(all_distances) == 0:
            return self.r, np.zeros_like(self.r)

        all_counts = [_dist_to_counts(d, r_min=self.r_min, r_max=self.r_max, n_grid=self.n_grid) for d in all_distances]
        sum_counts = np.sum(all_counts, axis=0)
        total_density = sum(density[i] for i in species)
        total_atoms = sum(n_atoms[i] for i in ref_species)
        rdf_temp = sum_counts / total_density / self.volumes / total_atoms
        if self.sigma > 1e-8:
            rdf_temp = gaussian_filter1d(rdf_temp, self.sigma)
        return self.r, rdf_temp

    def get_site_coordination(self, structure: Structure) -> tuple[np.ndarray, list[dict]]:
        """
        Get site wise coordination
        Args:
            structure (Structure): pymatgen Structure.

        Returns: r, cns where cns is a list of dictionary with specie_pair: pair_cn key:value pairs

        """
        _, rdfs = self.get_site_rdf(structure)
        density = self._get_specie_density(structure)
        cns: list[dict] = [{}] * len(structure)
        for i, rdf in enumerate(rdfs):
            if len(rdf) == 0:
                continue
            for pair, rdf_pair in rdf.items():
                _, specie = pair.split(":")
                cns[i][pair] = np.cumsum(rdf_pair * density[specie] * 4.0 * np.pi * self.r**2 * self.dr)
        return self.r, cns

    def get_species_coordination(
        self,
        structure: Structure,
        ref_species: list | None = None,
        species: list | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Get specie-wise coordination number
        Args:
            structure (Structure): target structure
            ref_species (list of species or just single specie str): the reference species.
                The rdfs are calculated with these species at the center
            species (list of species or just single specie str): the species that we are interested in.
                The rdfs are calculated on these species.

        Returns: 2-D array of distances and respective specie-wise coordination number.
        """
        all_species = list({str(i.specie) for i in structure.sites})
        if ref_species is None:
            ref_species = all_species
        if species is None:
            species = all_species

        _, rdf = self.get_species_rdf(structure=structure, ref_species=ref_species, species=species)

        density = self._get_specie_density(structure)
        n_atoms = structure.composition.to_data_dict["unit_cell_composition"]
        total_density = sum(density[i] for i in species)
        total_atoms = sum(n_atoms[i] for i in ref_species)
        return self.r, np.cumsum(rdf * total_density * 4.0 * np.pi * self.r**2 * self.dr * total_atoms)

    @staticmethod
    def _get_specie_density(structure: Structure):
        n_atoms = structure.composition.to_data_dict["unit_cell_composition"]
        density = {}
        for i, j in n_atoms.items():
            density[i] = j / structure.volume
        return density


def _dist_to_counts(d: np.ndarray, r_min: float = 0.0, r_max: float = 8.0, n_grid: int = 100) -> np.ndarray:
    """
    Convert a distance array for counts in the bin
    Args:
        d (1D np.ndarray): distance array
        r_min (float): minimal radius
        r_max (float): maximum radius
    Returns:
        1D array of counts in the bins centered on grid.
    """
    counts = np.zeros((n_grid,))
    dr = (r_max - r_min) / (n_grid - 1)  # end points are on grid
    indices = np.array(np.floor((d - r_min + 0.5 * dr) / dr), dtype=int)

    unique, val_counts = np.unique(indices, return_counts=True)
    counts[unique] = val_counts
    return counts


def get_pair_distances(structure: Structure, r_max: float = 8.0) -> list[dict]:
    """
    Get pair distances from structure.
    The output will be a list of dictionary, for example
    [{"specie": "Mo",
    "neighbors": {"S": [1.0, 2.0, ...], "Fe": [1.2, 3.0, ...]}},
    {"specie": "Fe",
    "neighbors": {"Mo": [1.0, 3.0, ...]}}]
    it will be fairly easy to construct radial distribution func, etc
    from here.

    Args:
        structure (Structure): pymatgen Structure
        r_max (float): maximum radius to consider

    Returns: a list of dictionary with pair distance information of each specie.
    """
    index1, index2, _, distances = structure.get_neighbor_list(r_max)
    species = np.array([str(i.specie) for i in structure.sites])
    res = [{"specie": i, "neighbors": {}} for i in species]
    neighbor_species = species[index2]
    tuples = np.array(list(zip(index1, neighbor_species, strict=False)), dtype=[("index", "i4"), ("specie", "<U10")])
    unique_tuples, indices = np.unique(tuples, return_inverse=True)
    for _index, unique_tuple in enumerate(unique_tuples):
        res[unique_tuple[0]]["neighbors"][unique_tuple[1]] = distances[tuples == unique_tuple]
    return res



"""MEGNet-based describers."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd

from maml.base import BaseDescriber, describer_type
from maml.utils import get_full_stats_and_funcs

if TYPE_CHECKING:
    from pymatgen.core import Molecule, Structure

DEFAULT_MODEL = Path(__file__).parent / "data/megnet_models/formation_energy.hdf5"


def _load_model(name: str | object | None = None):
    try:
        from megnet.utils.descriptor import MEGNetDescriptor
        from megnet.utils.models import MODEL_MAPPING, load_model
    except ImportError:
        raise MEGNetNotFound

    AVAILBLE_MODELS = list(MODEL_MAPPING.keys())
    if isinstance(name, str) and name in AVAILBLE_MODELS:
        name_or_model = load_model(name)
    elif name is None:
        name_or_model = str(DEFAULT_MODEL)

    else:
        name_or_model = name
    return MEGNetDescriptor(name_or_model)


@describer_type("site")
class MEGNetSite(BaseDescriber):
    """
    Use megnet pre-trained models as featurizer to get atomic features.

    Reference:
    @article{chen2019graph,title={Graph networks as a universal machine
                learning framework for molecules and crystals},
            author={Chen, Chi and Ye, Weike and Zuo, Yunxing and
                Zheng, Chen and Ong, Shyue Ping},
            journal={Chemistry of Materials}, volume={31}, number={9},
            pages={3564--3572}, year={2019},publisher={ACS Publications}}
    """

    def __init__(self, name: str | object | None = None, level: int | None = None, **kwargs):
        """

        Args:
            name (str or megnet.models.GraphModel): models name keys, megnet models
                path or a MEGNet GraphModel, if no name is provided, the models will be Eform_MP_2019.
            level (int): megnet graph layer level.
            **kwargs: passthrough.
        """
        self.describer_model = _load_model(name)

        if level is None:
            n_layers = sum(i.startswith("meg_net") for i in self.describer_model.valid_names) // 3
            level = n_layers
        self.name = name
        self.level = level
        super().__init__(**kwargs)

    def transform_one(self, obj: Structure | Molecule):
        """
        Get megnet site features from structure object.

        Args:
            obj (structure or molecule): pymatgen structure or molecules

        Returns: A pandas dataframe of MEGNet atom features.

        """
        features = self.describer_model.get_atom_features(obj, level=self.level)
        return pd.DataFrame(features)

    def __getstate__(self):
        """
        Get state for pickle
        Returns: dictionary.
        """
        d = self.__dict__.copy()
        d["describer_model"] = None
        return d

    def __setstate__(self, state):
        """
        Set state of object.

        Args:
            state: a dict of all attributes of the MEGNetSite.

        Returns: None.
        """
        for i, j in state.items():
            setattr(self, i, j)
        self.describer_model = _load_model(self.name)


@describer_type("structure")
class MEGNetStructure(BaseDescriber):
    """
    Use megnet pre-trained models as featurizer to get
    structural features. There are two methods to get structural descriptors from
    megnet models.

    mode:
        'site_stats': Calculate the site features, and then use maml.utils.stats to compute the feature-wise
            statistics. This requires the specification of level
        'site_readout': Use the atomic features at the readout stage
        'final': Use the concatenated atom, bond and global features

    Reference:
    @article{chen2019graph,title={Graph networks as a universal machine
                learning framework for molecules and crystals},
            author={Chen, Chi and Ye, Weike and Zuo, Yunxing and
                Zheng, Chen and Ong, Shyue Ping},
            journal={Chemistry of Materials}, volume={31}, number={9},
            pages={3564--3572}, year={2019},publisher={ACS Publications}}
    """

    def __init__(
        self,
        name: str | object | None = None,
        mode: str = "site_stats",
        level: int | None = None,
        stats: list | None = None,
        **kwargs,
    ):
        """

        Args:s
            name (str or megnet.models.GraphModel): models name keys, megnet models path or
                a MEGNet GraphModel, if no name is provided, the models will be Eform_MP_2019.
            mode (str): choose one from ['site_stats', 'site_readout', 'final'].
                'site_stats': Calculate the site features, and then use maml.utils.stats to compute the feature-wise
                    statistics. This requires the specification of level
                'site_readout': Use the atomic features at the readout stage
                'state': Use the state attributes
                'final': Use the concatenated atom, bond and global features
            level (int): megnet graph layer level.
            **kwargs: passthrough.
        """
        self.describer_model = _load_model(name)

        if level is None:
            n_layers = sum(i.startswith(("meg_net", "megnet")) for i in self.describer_model.valid_names) // 3
            level = n_layers

        self.name = name
        self.level = level
        self.mode = mode
        if stats is None:
            stats = ["min", "max", "range", "mean", "mean_absolute_error", "mode"]
        self.stats = stats
        full_stats, stats_func = get_full_stats_and_funcs(stats)
        self.full_stats = full_stats
        self.stats_func = stats_func
        super().__init__(**kwargs)

    def transform_one(self, obj: Structure | Molecule):
        """
        Transform structure/molecule objects into features
        Args:
            obj (Structure/Molecule): target object structure or molecule.

        Returns: pd.DataFrame features

        """
        if self.mode == "site_stats":
            features = self.describer_model.get_atom_features(obj, level=self.level)
            features_transpose = list(zip(*features, strict=False))
            column_names = []
            final_features = []
            for i, f in enumerate(features_transpose):
                column_names.extend([f"{i}_{n}" for n in self.full_stats])
                final_features.extend([func(f) for func in self.stats_func])
            return pd.DataFrame([final_features], columns=column_names)

        if self.mode == "site_readout":
            return pd.DataFrame(self.describer_model.get_set2set(obj, ftype="atom"))
        if self.mode == "state":
            return pd.DataFrame(self.describer_model.get_global_features(obj, level=self.level))
        if self.mode == "final":
            return pd.DataFrame(self.describer_model.get_structure_features(obj))
        raise ValueError("Mode not allowed.")

    def __getstate__(self):
        """
        Get state for pickle.
        Returns: dictionary.
        """
        d = self.__dict__.copy()
        d["describer_model"] = None
        return d

    def __setstate__(self, state):
        """
        Set state of object.

        Args:
            state: a dict of attributes of MEGNetStructure.

        Returns: None.
        """
        for i, j in state.items():
            setattr(self, i, j)
        self.describer_model = _load_model(self.name)


class MEGNetNotFound(Exception):
    """MEGNet not found exception."""

    def __init__(self):
        """MEGNet not found exception."""
        super().__init__(
            "This module requires installation of megnet, "
            "which is not found in your current environment."
            "Please install it via `pip install megnet` or "
            "via github source"
        )



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""Define abstract base classes."""
from __future__ import annotations

from ._data import BaseDataSource
from ._describer import BaseDescriber, DummyDescriber, SequentialDescriber, describer_type
from ._feature_batch import get_feature_batch
from ._mixin import TargetScalerMixin
from ._model import BaseModel, KerasModel, SKLModel, is_keras_model, is_sklearn_model

__all__ = [
    "BaseDataSource",
    "BaseDescriber",
    "BaseModel",
    "DummyDescriber",
    "KerasModel",
    "SKLModel",
    "SequentialDescriber",
    "TargetScalerMixin",
    "describer_type",
    "get_feature_batch",
    "is_keras_model",
    "is_sklearn_model",
]



"""maml - materials machine learning."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("maml")
except PackageNotFoundError:
    pass  # package not installed



"""
Temporary directory and file creation utilities.
This file is adapted from monty.tempfile.
"""

from __future__ import annotations

import os
import shutil
import tempfile
import warnings
from pathlib import Path
from typing import Any

from monty.shutil import copy_r, remove


class MultiScratchDir:
    """
    Automatically handles creation of temporary directories (utilizing Python's build in temp directory functions).

    The main difference between this class and monty ScratchDir is that multiple temp directories are created here.
    It enables the running of multiple jobs simultaneously in the directories
    The way it works is as follows:

    1. Create multiple temp dirs in specified root path.
    2. Optionally copy input files from current directory to temp dir.
    3. User loops among all directories
    4. User performs specified operations in each directories
    5. Change back to original directory.
    6. Delete temp dir.
    """

    SCR_LINK = "scratch_link"

    def __init__(
        self,
        rootpath: str | Path,
        n_dirs: int = 1,
        create_symbolic_link: bool = False,
        copy_from_current_on_enter: bool = False,
        copy_to_current_on_exit: bool = False,
    ):
        """
        Initializes scratch directory given a **root** path. There is no need
        to try to create unique directory names. The code will generate a
        temporary sub directory in the rootpath. The way to use this is using a
        with context manager. Example::
            with ScratchDir("/scratch"):
                do_something()
        If the root path does not exist or is None, this will function as a
        simple pass through, i.e., nothing happens.

        Args:
            rootpath (str/Path): Path in which to create temp subdirectories.
                If this is None, no temp directories will be created and
                this will just be a simple pass through.
            n_dirs (int): number of temporary directories to create
            create_symbolic_link (bool): Whether to create a symbolic link in
                the current working directory to the scratch directory
                created.
            copy_from_current_on_enter (bool): Whether to copy all files from
                the current directory (recursively) to the temp dir at the
                start, e.g., if input files are needed for performing some
                actions. Defaults to False.
            copy_to_current_on_exit (bool): Whether to copy files from the
                scratch to the current directory (recursively) at the end. E
                .g., if output files are generated during the operation.
                Defaults to False.
        """
        if Path is not None and isinstance(rootpath, Path):
            rootpath = str(rootpath)

        self.rootpath = os.path.abspath(rootpath) if rootpath is not None else None
        self.n_dirs = n_dirs
        self.cwd = os.getcwd()
        self.create_symbolic_link = create_symbolic_link
        self.start_copy = copy_from_current_on_enter
        self.end_copy = copy_to_current_on_exit
        self.tempdirs: list[str] = []

    def __enter__(self):
        tempdirs = [self.cwd] * self.n_dirs
        if self.rootpath is not None and os.path.exists(self.rootpath):
            tempdirs = [tempfile.mkdtemp(dir=self.rootpath) for _ in range(self.n_dirs)]
            self.tempdirs = [os.path.abspath(tempdir) for tempdir in tempdirs]
            if self.start_copy:
                [copy_r(".", tempdir) for tempdir in tempdirs]
            if self.create_symbolic_link:
                [os.symlink(tempdir, f"{MultiScratchDir.SCR_LINK}_{i}") for i, tempdir in enumerate(tempdirs)]
        return tempdirs

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object):
        if self.rootpath is not None and os.path.exists(self.rootpath):
            if self.end_copy:
                # First copy files over
                [_copy_r_with_suffix(tempdir, self.cwd, i) for i, tempdir in enumerate(self.tempdirs)]

            os.chdir(self.cwd)
            for tempdir in self.tempdirs:
                remove(tempdir)


def _copy_r_with_suffix(src: str, dst: str, suffix: Any | None = None):
    """
    Implements a recursive copy function similar to Unix's "cp -r" command.
    Surprisingly, python does not have a real equivalent. shutil.copytree
    only works if the destination directory is not present.

    Args:
        src (str): Source folder to copy.
        dst (str): Destination folder.
        suffix: Suffix to be added for copied files.
    """
    abssrc = os.path.abspath(src)
    absdst = os.path.abspath(dst)

    try:
        os.makedirs(absdst)
    except OSError:
        # If absdst exists, an OSError is raised. We ignore this error.
        pass
    for f in os.listdir(abssrc):
        fpath = os.path.join(abssrc, f)
        if os.path.isfile(fpath):
            if suffix is not None:
                new_path = f"{fpath}_{suffix!s}"
                shutil.copy(fpath, new_path)
                fpath = new_path
            shutil.copy(fpath, absdst)

        elif not absdst.startswith(fpath):
            _copy_r_with_suffix(fpath, os.path.join(absdst, f), suffix=suffix)
        else:
            warnings.warn(f"Cannot copy {fpath} to itself")



"""DFT wrapper."""

from __future__ import annotations

import os
import subprocess
from shutil import which
from typing import TYPE_CHECKING

from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.entries.compatibility import MaterialsProjectCompatibility
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.io.vasp.sets import MPStaticSet

from maml.apps.bowsr.model.base import EnergyModel

if TYPE_CHECKING:
    from pymatgen.core.structure import Structure

module_dir = os.path.dirname(__file__)
elements_filename = os.path.join(module_dir, "..", "regularization", "elements.json")
elements = loadfn(elements_filename)


class DFT(EnergyModel):
    """DFT static calculation wrapped as energy model."""

    def __init__(self, exe_path: str | None = None):
        """
        DFT wrapper
        Args:
            exe_path: VASP executable path.
        """
        if not exe_path:
            if not which("vasp_std"):
                raise RuntimeError("Vasp executable can not be found.")
            self.vasp_exe = which("vasp_std")
        else:
            self.vasp_exe = exe_path

    def predict_energy(self, structure: Structure):
        """
        Predict energy from structure.

        Args:
            structure: (pymatgen Structure).

        Returns: float
        """
        el_amt_dict = structure.composition.get_el_amt_dict()

        static = MPStaticSet(structure)
        with ScratchDir("."):
            static.write_input(".")
            with subprocess.Popen(
                [self.vasp_exe], stdout=subprocess.PIPE, stderr=subprocess.PIPE  # type: ignore
            ) as p_exe:
                stdout, stderr = p_exe.communicate()
                rc = p_exe.returncode
            if rc != 0:
                error_msg = f"vasp exited with return code {rc}"
                msg = stderr.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += ", "
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            compat = MaterialsProjectCompatibility()
            vrun = Vasprun("vasprun.xml")
            entry = compat.process_entry(vrun.get_computed_entry())
            return (
                entry.energy  # type: ignore[union-attr]
                - sum(elements[el]["energy_per_atom"] * amt for el, amt in el_amt_dict.items())
            ) / len(structure)



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides SOAP-GAP interatomic potential class."""
from __future__ import annotations

import os
import re
import subprocess
import xml.etree.ElementTree as ET
from collections import OrderedDict, defaultdict
from shutil import which

import numpy as np
from monty.io import zopen
from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.core import Element, Lattice, Structure
from pymatgen.core.periodic_table import get_el_sp
from ruamel import yaml

from maml.utils import check_structures_forces_stresses, convert_docs, pool_from

from ._lammps import LammpsPotential

module_dir = os.path.dirname(__file__)
soap_params = loadfn(os.path.join(module_dir, "params", "GAP.json"))


class GAPotential(LammpsPotential):
    """This class implements Smooth Overlap of Atomic Position potentials."""

    pair_style = "pair_style        quip"
    pair_coeff = "pair_coeff        * * {} {} {}"

    def __init__(self, name=None, param=None):
        """

        Args:
            name (str): Name of force field.
            param (dict): The parameter configuration of potentials.
        """
        self.name = name if name else "GAPotential"
        self.param = param if param else {}
        self.elements = None

    @staticmethod
    def _line_up(structure, energy, forces, virial_stress):
        """
        Convert input structure, energy, forces, virial_stress to
        proper configuration format for MLIP usage.

        Args:
            structure (Structure): Pymatgen Structure object.
            energy (float): DFT-calculated energy of the system.
            forces (list): The forces should have dimension
                (num_atoms, 3).
            virial_stress (list): stress should has 6 distinct
                elements arranged in order [xx, yy, zz, xy, yz, xz].

        Returns:
            outputs: a proper format storing structure, energy, forces, virial_stress.
        """
        full_virial_stress = [
            virial_stress[0],
            virial_stress[3],
            virial_stress[5],
            virial_stress[3],
            virial_stress[1],
            virial_stress[4],
            virial_stress[5],
            virial_stress[4],
            virial_stress[2],
        ]

        inputs = OrderedDict(
            Size=structure.num_sites,
            SuperCell=structure.lattice,
            AtomData=(structure, forces),
            Energy=energy,
            Stress=full_virial_stress,
        )

        lines = []
        if "Size" in inputs:
            lines.append(str(inputs["Size"]))

        description = []
        if "Energy" in inputs:
            description.append("dft_energy=" + str(inputs["Energy"]))
        if "Stress" in inputs:
            description.append("dft_virial=" + "\t".join([str(f) for f in inputs["Stress"]]))
        if "SuperCell" in inputs:
            super_cell_str = [str(f) for f in inputs["SuperCell"].matrix.ravel()]
            description.append(f'Lattice="{"     ".join(super_cell_str)}"')
        description.append("Properties=species:S:1:pos:R:3:Z:I:1:dft_force:R:3")
        lines.append(" ".join(description))

        if "AtomData" in inputs:
            format_str = "{:<10s}{:>16f}{:>16f}{:>16f}{:>8d}{:>16f}{:>16f}{:>16f}"
            for _i, (site, force) in enumerate(zip(structure, forces, strict=False)):
                lines.append(format_str.format(site.species_string, *site.coords, site.specie.Z, *force))
        return "\n".join(lines)

    def write_cfgs(self, filename, cfg_pool):
        """
        Write the formatted configuration file.

        Args:
            filename (str): The filename to be written.
            cfg_pool (list): The configuration pool contains
                structure and energy/forces properties.
        """
        if not filename.endswith(".xyz"):
            raise RuntimeError('The extended xyz file should end with ".xyz"')

        lines = []
        for dataset in cfg_pool:
            if isinstance(dataset["structure"], dict):
                structure = Structure.from_dict(dataset["structure"])
            else:
                structure = dataset["structure"]
            energy = dataset["outputs"]["energy"]
            forces = dataset["outputs"]["forces"]
            virial_stress = dataset["outputs"]["virial_stress"]

            lines.append(self._line_up(structure, energy, forces, virial_stress))

        with open(filename, "w") as f:
            f.write("\n".join(lines))

        return filename

    @staticmethod
    def read_cfgs(filename, predict=False):
        """
        Read the configuration file.

        Args:
            filename (str): The configuration file to be read.
            predict (bool): whether it is the prediction mode or not.
        """
        type_convert = {"R": np.float32, "I": np.int64, "S": np.str_}
        data_pool = []
        with zopen(filename, "rt") as f:
            lines = f.read()
        repl = re.compile("AT ")
        lines = repl.sub("", string=lines)

        block_pattern = re.compile(r"(\n[0-9]+\n|^[0-9]+\n)(.+?)(?=\n[0-9]+\n|$)", re.S)
        lattice_pattern = re.compile(r'Lattice="(.+)"')
        # energy_pattern = re.compile('dft_energy=(-?[0-9]+.[0-9]+)', re.I)
        energy_pattern = re.compile(r"(?<=\S{3}\s|dft_)energy=(-?[0-9]+.[0-9]+)")
        # stress_pattern = re.compile('dft_virial={(.+)}')
        stress_pattern = re.compile(r"dft_virial=({|)(.+?)(}|) \S.*")
        properties_pattern = re.compile(r"properties=(\S+)", re.I)
        # position_pattern = re.compile('\n(.+)', re.S)
        position_pattern = re.compile("\n(.+?)(?=\nE.*|\n\n.*|$)", re.S)
        # formatify = lambda string: [float(s) for s in string.split()]

        for size, block in block_pattern.findall(lines):
            d = {"outputs": {}}
            size = int(size)
            lattice_str = lattice_pattern.findall(block)[0]
            lattice = Lattice([float(s) for s in lattice_str.split()])
            energy_str = energy_pattern.findall(block)[-1]
            energy = float(energy_str)
            stress_str = stress_pattern.findall(block)[0][1]
            virial_stress = np.array([float(s) for s in stress_str.split()])
            virial_stress = [virial_stress[i] for i in [0, 4, 8, 1, 5, 6]]
            properties = properties_pattern.findall(block)[0].split(":")
            labels_columns = OrderedDict()
            labels = defaultdict()
            for i in range(0, len(properties), 3):
                labels_columns[properties[i]] = [int(properties[i + 2]), properties[i + 1]]
            position_str = position_pattern.findall(block)[0].split("\n")
            position = np.array([p.split() for p in position_str])
            column_index = 0
            for key in labels_columns:
                num_columns, dtype = labels_columns[key]
                labels[key] = position[:, column_index : column_index + num_columns].astype(type_convert[dtype])
                column_index += num_columns
            struct = Structure(
                lattice=lattice, species=labels["species"].ravel(), coords=labels["pos"], coords_are_cartesian=True
            )
            forces = labels["force"] if predict else labels["dft_force"]
            d["structure"] = struct.as_dict()
            d["outputs"]["energy"] = energy
            assert size == struct.num_sites
            d["num_atoms"] = size
            d["outputs"]["forces"] = forces
            d["outputs"]["virial_stress"] = virial_stress

            data_pool.append(d)
        _, df = convert_docs(docs=data_pool)
        return data_pool, df

    def train(
        self,
        train_structures,
        train_energies,
        train_forces,
        train_stresses=None,
        default_sigma=(0.0005, 0.1, 0.05, 0.01),
        use_energies=True,
        use_forces=True,
        use_stress=False,
        **kwargs,
    ):
        """
        Training data with gaussian process regression.

        Args:
            train_structures ([Structure]): The list of Pymatgen Structure object.
                energies ([float]): The list of total energies of each structure
                in structures list.
            train_energies ([float]): List of total energies of each structure in
                structures list.
            train_forces ([np.array]): List of (m, 3) forces array of each structure
                with m atoms in structures list. m can be varied with each
                single structure case.
            train_stresses (list): List of (6, ) virial stresses of each
                structure in structures list.
            default_sigma (list): Error criteria in energies, forces, stress
                and hessian. Should have 4 numbers.
            use_energies (bool): Whether to use dft total energies for training.
                Default to True.
            use_forces (bool): Whether to use dft atomic forces for training.
                Default to True.
            use_stress (bool): Whether to use dft virial stress for training.
                Default to False.

            kwargs:
                l_max (int): Parameter to configure GAP. The band limit of
                    spherical harmonics basis function. Default to 12.
                n_max (int): Parameter to configure GAP. The number of radial basis
                    function. Default to 10.
                atom_sigma (float): Parameter to configure GAP. The width of gaussian
                    atomic density. Default to 0.5.
                zeta (float): Present when covariance function type is do product.
                    Default to 4.
                cutoff (float): Parameter to configure GAP. The cutoff radius.
                    Default to 4.0.
                cutoff_transition_width (float): Parameter to configure GAP.
                    The transition width of cutoff radial. Default to 0.5.
                delta (float): Parameter to configure Sparsification.
                    The signal variance of noise. Default to 1.
                f0 (float): Parameter to configure Sparsification.
                    The signal mean of noise. Default to 0.0.
                n_sparse (int): Parameter to configure Sparsification.
                    Number of sparse points.
                covariance_type (str): Parameter to configure Sparsification.
                    The type of convariance function. Default to dot_product.
                sparse_method (str): Method to perform clustering in sparsification.
                    Default to 'cur_points'.

                sparse_jitter (float): Intrisic error of atomic/bond energy,
                    used to regularise the sparse covariance matrix.
                    Default to 1e-8.
                e0 (float): Atomic energy value to be subtracted from energies
                    before fitting. Default to 0.0.
                e0_offset (float): Offset of baseline. If zero, the offset is
                    the average atomic energy of the input data or the e0
                    specified manually. Default to 0.0.
        """
        if not which("gap_fit"):
            raise RuntimeError(
                "gap_fit has not been found.\n",
                "Please refer to https://github.com/libAtoms/QUIP for ",
                "further detail.",
            )

        train_structures, train_forces, train_stresses = check_structures_forces_stresses(
            train_structures, train_forces, train_stresses
        )

        gap_sorted_elements = []
        for struct in train_structures:
            for specie in struct.species:
                if str(specie) not in gap_sorted_elements:
                    gap_sorted_elements.append(str(specie))

        self.elements = sorted(gap_sorted_elements, key=lambda x: Element(x))

        atoms_filename = "train.xyz"
        xml_filename = "train.xml"
        train_pool = pool_from(train_structures, train_energies, train_forces, train_stresses)

        exe_command = ["gap_fit"]
        exe_command.append(f"at_file={atoms_filename}")
        gap_configure_params = [
            "l_max",
            "n_max",
            "atom_sigma",
            "zeta",
            "cutoff",
            "cutoff_transition_width",
            "delta",
            "f0",
            "n_sparse",
            "covariance_type",
            "sparse_method",
        ]
        preprocess_params = ["sparse_jitter", "e0", "e0_offset"]
        if len(default_sigma) != 4:
            raise ValueError("The default sigma is supposed to have 4 numbers.")

        gap_command = ["soap"]
        for param_name in gap_configure_params:
            param = kwargs.get(param_name) if kwargs.get(param_name) else soap_params.get(param_name)
            gap_command.append(param_name + "=" + f"{param}")
        gap_command.append("add_species=T")
        exe_command.append("gap={" + " ".join(gap_command) + "}")

        for param_name in preprocess_params:
            param = kwargs.get(param_name) if kwargs.get(param_name) else soap_params.get(param_name)
            exe_command.append(param_name + "=" + f"{param}")

        default_sigma = [str(f) for f in default_sigma]
        exe_command.append(f"default_sigma={{{' '.join(default_sigma)}}}")

        if use_energies:
            exe_command.append("energy_parameter_name=dft_energy")
        if use_forces:
            exe_command.append("force_parameter_name=dft_force")
        if use_stress:
            exe_command.append("virial_parameter_name=dft_virial")
        exe_command.append(f"gp_file={xml_filename}")

        with ScratchDir("."):
            self.write_cfgs(filename=atoms_filename, cfg_pool=train_pool)

            with subprocess.Popen(exe_command, stdout=subprocess.PIPE) as p:
                stdout = p.communicate()[0]
                rc = p.returncode
            if rc != 0:
                error_msg = f"gap_fit exited with return code {rc}"
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            def get_xml(xml_file):
                tree = ET.parse(xml_file)
                root = tree.getroot()
                potential_label = root.tag
                element_param = {}
                for gpcoordinates in list(root.iter("gpCoordinates")):
                    gp_descriptor = next(iter(gpcoordinates.iter("descriptor")))
                    gp_descriptor_text = gp_descriptor.findtext(".")
                    Z_pattern = re.compile(" Z=(.*?) ", re.S)
                    element = str(get_el_sp(int(Z_pattern.findall(gp_descriptor_text)[0])))
                    param = np.loadtxt(gpcoordinates.get("sparseX_filename"))
                    element_param[element] = param.tolist()

                return tree, element_param, potential_label

            tree, element_param, potential_label = get_xml(xml_filename)
            self.param["xml"] = tree
            self.param["element_param"] = element_param
            self.param["potential_label"] = potential_label

        return rc

    def write_param(self, xml_filename="gap.2020.01.xml"):
        """
        Write xml file to perform lammps calculation.

        Args:
            xml_filename (str): Filename to store xml formatted parameters.
        """
        if not self.param:
            raise RuntimeError("The xml and parameters should be provided.")
        tree = self.param.get("xml")
        root = tree.getroot()
        element_param = self.param.get("element_param")
        atomic_numbers = []
        for _i, gpcoordinates in enumerate(list(root.iter("gpCoordinates"))):
            gp_descriptor = next(iter(gpcoordinates.iter("descriptor")))
            gp_descriptor_text = gp_descriptor.findtext(".")
            Z_pattern = re.compile(" Z=(.*?) ", re.S)
            element = str(get_el_sp(int(Z_pattern.findall(gp_descriptor_text)[0])))
            atomic_numbers.append(str(Element(element).number))
            param_file = f"{element}.soapparam"
            gpcoordinates.set("sparseX_filename", param_file)
            np.savetxt(param_file, element_param[element], fmt="%.20e")
        tree.write(xml_filename)

        pair_coeff = self.pair_coeff.format(
            xml_filename, f'"Potential xml_label={self.param.get("potential_label")}"', " ".join(atomic_numbers)
        )
        return [self.pair_style, pair_coeff]

    def evaluate(
        self,
        test_structures,
        test_energies,
        test_forces,
        test_stresses=None,
        predict_energies=True,
        predict_forces=True,
        predict_stress=False,
    ):
        """
        Evaluate energies, forces and stresses of structures with trained
        interatomic potentials.

        Args:
            test_structures ([Structure]): List of Pymatgen Structure Objects.
            test_energies ([float]): List of DFT-calculated total energies of
                each structure in structures list.
            test_forces ([np.array]): List of DFT-calculated (m, 3) forces of
                each structure with m atoms in structures list. m can be varied
                with each single structure case.
            test_stresses (list): List of DFT-calculated (6, ) viriral stresses
                of each structure in structures list.
            predict_energies (bool): Whether to predict energies of configurations.
            predict_forces (bool): Whether to predict forces of configurations.
            predict_stress (bool): Whether to predict virial stress of
                configurations.
        """
        test_structures, test_forces, test_stresses = check_structures_forces_stresses(
            test_structures, test_forces, test_stresses
        )
        if not which("quip"):
            raise RuntimeError(
                "quip has not been found.\n", "Please refer to https://github.com/libAtoms/QUIP for ", "further detail."
            )
        xml_file = "predict.xml"
        original_file = "original.xyz"
        predict_file = "predict.xyz"
        predict_pool = pool_from(test_structures, test_energies, test_forces, test_stresses)

        with ScratchDir("."):
            _ = self.write_param(xml_file)
            original_file = self.write_cfgs(original_file, cfg_pool=predict_pool)
            _, df_orig = self.read_cfgs(original_file)

            exe_command = ["quip"]
            exe_command.append(f"atoms_filename={original_file}")
            exe_command.append(f"param_filename={xml_file}")
            if predict_energies:
                exe_command.append("energy=T")
            if predict_forces:
                exe_command.append("forces=T")
            if predict_stress:
                exe_command.append("virial=T")

            with open(predict_file, "w") as f, subprocess.Popen(exe_command, stdout=f) as p:
                _ = p.communicate()[0]

            _, df_predict = self.read_cfgs(predict_file, predict=True)

        return df_orig, df_predict

    def save(self, filename="param.yaml"):
        """
        Save parameters of the potentials.

        Args:
            filename (str): The file to store parameters of potentials.

        Returns:
            (str)
        """
        with open(filename, "w") as f:
            yaml.dump(self.param, f)

        return filename

    @staticmethod
    def from_config(filename):
        """
        Initialize potentials with parameters file.

        Args:
            filename (str): The file storing parameters of potentials,
                filename should ends with ".xml".

        Returns:
            GAPotential.
        """

        def get_xml(xml_file):
            tree = ET.parse(xml_file)
            root = tree.getroot()
            potential_label = root.tag
            element_param = {}

            for gpcoordinates in list(root.iter("gpCoordinates")):
                gp_descriptor = next(iter(gpcoordinates.iter("descriptor")))
                gp_descriptor_text = gp_descriptor.findtext(".")
                Z_pattern = re.compile(" Z=(.*?) ", re.S)
                element = str(get_el_sp(int(Z_pattern.findall(gp_descriptor_text)[0])))
                param = np.loadtxt(gpcoordinates.get("sparseX_filename"))
                element_param[element] = param.tolist()

            return tree, element_param, potential_label

        tree, element_param, potential_label = get_xml(filename)
        parameters = dict(xml=tree, element_param=element_param, potential_label=potential_label)
        gap = GAPotential(param=parameters)
        gap.elements = sorted(element_param.keys(), key=lambda x: Element(x))
        return gap



"""Batch a list of features output by describers.transform method."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from collections.abc import Callable


def pandas_concat(features: list[pd.DataFrame]) -> pd.DataFrame:
    """
    Concatenate a list of pandas dataframe into a single one
    Args:
        features (list): list of pandas dataframe.

    Returns: concatenated pandas dataframe

    """
    return pd.concat(features, keys=range(len(features)), names=["input_index", None])


def stack_first_dim(features: list[np.ndarray]) -> np.ndarray:
    """
    Stack the first dimension. If the original features
    are a list of nxm array, the stacked features will be lxnxm,
    where l is the number of entries in the list
    Args:
        features (list): list of numpy array features.

    Returns: stacked features

    """
    return np.stack(features)


def stack_padded(features: list[np.ndarray]) -> np.ndarray:
    """
    Stack the first dimension. If the original features
    are a list of nxm array, the stacked features will be lxnxm,
    where l is the number of entries in the list
    Args:
        features (list): list of numpy array features.

    Returns: stacked features

    """
    feature_dims = np.array([i.shape[0] for i in features])
    n_data = len(feature_dims)
    max_feature_dims = np.max(feature_dims)
    res = np.zeros((n_data, max_feature_dims))
    index1 = np.tile(np.arange(max_feature_dims)[None, :], (n_data, 1))
    masks = index1 < feature_dims[:, None]
    res[masks] = np.concatenate(features)
    return res


def no_action(features: list[Any]) -> list[Any]:
    """Return original feature lists."""
    return features


AVAILABLE_FB_METHODS = {
    "pandas_concat": pandas_concat,
    "stack_first_dim": stack_first_dim,
    "stack_padded": stack_padded,
    "no_action": no_action,
}


def get_feature_batch(fb_name: str | Callable | None = None) -> Callable:
    """
    Providing a feature batch name, returning the function callable
    Args:
        fb_name (str): name of the feature batch function
    Returns: callable feature batch function.
    """
    if fb_name is None:
        return no_action

    if isinstance(fb_name, str):
        try:
            return AVAILABLE_FB_METHODS[fb_name]
        except KeyError:
            raise KeyError("Feature batch method not supported! Available ones are " + str(AVAILABLE_FB_METHODS.keys()))
    else:
        return fb_name



from __future__ import annotations

import os
import unittest

import numpy as np
from pymatgen.core import Structure
from pymatgen.util.testing import PymatgenTest

from maml.apps.bowsr.model.cgcnn import CGCNN, torch

try:
    import cgcnn
except ImportError:
    cgcnn = None

module_dir = os.path.dirname(__file__)
pjoin = os.path.join


class TestCGCNN(PymatgenTest):
    @unittest.skipIf(cgcnn is None or torch is None, "cgcnn or torch is not installed")
    def test_predict_energy(self):
        structure = Structure.from_file(pjoin(module_dir, "mp-6812.cif"))
        cgcnn = CGCNN()
        output = cgcnn.predict_energy(structure)
        self.assertAlmostEqual(output.astype(np.float64), -3.0786154)


if __name__ == "__main__":
    unittest.main()



"""Module implements helper functions to retrieve data for GB energy prediction paper."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from monty.serialization import dumpfn, loadfn
from pymatgen.ext.matproj import MPRester, MPRestError

from maml.apps import gbe

if TYPE_CHECKING:
    from pymatgen.core import Structure

pjoin = os.path.join
module_dir = os.path.dirname(gbe.__file__)
REFS = pjoin(module_dir, "references")
DATA = pjoin(module_dir, "data")


def load_data(filename: str | None = None) -> list:
    """
    Helper function to load the data
    Default is to load the 361 data
    Args:
        filename (str): the filename of the data.

    Returns:
        data list

    """
    if filename:
        return loadfn(filename)
    return loadfn(pjoin(DATA, "gb_data.json"))


def load_b0_dict() -> dict:
    """
    Helper function to retrieve the b0 data.
    b0 is the bulk bond length.

    Returns:
        returns the dict of {element: b0}
    """
    if os.path.isfile(pjoin(REFS, "el2b0.json")):
        return loadfn(pjoin(REFS, "el2b0.json"))
    print("Downloading Data ... ")
    if os.getenv("MAPI_KEY"):
        data = load_data()
        b0_dict = update_b0_dict(data, mp_api=os.getenv("MAPI_KEY"))
        dumpfn(b0_dict, pjoin(REFS, "el2b0.json"))
        return b0_dict
    raise MPRestError("No MAPI_KEY found")


def update_b0_dict(data_source: list, mp_api: str | None) -> dict:
    """
    Helper function to update the b0 dictionary
    Requires api key to MP
    Args:
        data_source (list): list of GB data
        mp_api (str): API key to MP
    Returns:
         b0_dict (dict): {el: b0}.
    """
    rester = MPRester(mp_api)

    def get_b0(bulk_structure: Structure) -> float:
        """b0 is the bond length of bulk metal."""
        _b0 = bulk_structure.get_distance(0, 0, jimage=1)
        return min(_b0, bulk_structure.lattice.a)

    b0_dict = {}
    for d in data_source:
        if d["pretty_formula"] in b0_dict:
            continue
        bs = rester.get_structure_by_material_id(d["material_id"])  # type: ignore[attr-defined]
        b0 = get_b0(bs)
        b0_dict.update({d["pretty_formula"]: b0})
    return b0_dict


def load_mean_delta_bl_dict(loc_algo: str = "crystalnn") -> dict:
    """
    Helper function to load the mean_delta_bl data
    Args:
        loc_algo (str): name of the algorithm.

    Returns:
        {task_id (int): mean_delta_bl(float)}

    """
    if os.path.isfile(pjoin(REFS, f"mean_bl_chg_{loc_algo}.json")):
        return loadfn(pjoin(REFS, f"mean_bl_chg_{loc_algo}.json"))
    raise ValueError(
        "Please provide mean_delta_bond_length data. "
        "Use gbe.describers.GBond.get_mean_bl_chg method"
        "to calculate mean_delta_bond_length"
    )



"""Module implements the scaler."""

from __future__ import annotations

import numpy as np
from monty.json import MSONable


class StandardScaler(MSONable):
    """
    StandardScaler follows the sklean manner with addition of
    dictionary representation.
    """

    def __init__(
        self,
        mean: list | np.ndarray | None = None,
        std: list | np.ndarray | None = None,
    ):
        """
        Args:
            mean: np.ndarray, mean values
            std: np.ndnarray, standard deviations.
        """
        self.mean = mean
        self.std = std

    def fit(self, target: list | np.ndarray) -> None:
        """
        Fit the StandardScaler to the target.

        Args:
            target (ndarray): The (mxn) ndarray. m is the number of samples,
                n is the number of feature dimensions.
        """
        mean = np.mean(target, axis=0)
        std = np.std(target, axis=0)
        self.mean = mean
        self.std = std

    def transform(self, target: np.ndarray) -> np.ndarray:
        """
        Transform target according to the mean and std.

        Args:
            target (ndarray): The (mxn) ndarray. m is the number of samples,
                n is the number of feature dimensions.
        """
        if self.mean is None or self.std is None:
            raise ValueError("No parameters is given.")
        return (target - self.mean) / self.std

    def inverse_transform(self, transformed_target: np.ndarray) -> np.ndarray:
        """
        Inversely transform the target.

        Args:
            transformed_target (ndarray): The (mxn) ndarray. m is the number of samples,
                n is the number of feature dimensions.
        """
        if self.mean is None or self.std is None:
            raise ValueError("No parameters is given.")
        return transformed_target * self.std + self.mean

    def __repr__(self):
        return f"{self.__class__.__name__}(mean={self.mean}, std={self.std})"

    def as_dict(self):
        """Dict representation of StandardScaler."""
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "params": {"mean": self.mean.tolist(), "std": self.std.tolist()},
        }

    @classmethod
    def from_dict(cls, d):
        """
        Reconstitute a StandardScaler object from a dict representation of
        StandardScaler created using as_dict().

        Args:
            d (dict): Dict representation of StandardScaler.
        """
        return cls(**d["params"])


class DummyScaler(MSONable):
    """Dummy scaler does nothing."""

    def fit(self, target: list | np.ndarray) -> None:
        """
        Fit the DummyScaler to the target.

        Args:
            target (ndarray): The (mxn) ndarray. m is the number of samples,
                n is the number of feature dimensions.
        """
        return

    def transform(self, target: list | np.ndarray) -> list | np.ndarray:
        """
        Transform target.

        Args:
            target (ndarray): The (mxn) ndarray. m is the number of samples,
                n is the number of feature dimensions.
        """
        return target

    def inverse_transform(self, transformed_target: list | np.ndarray) -> list | np.ndarray:
        """
        Inversely transform the target.

        Args:
            transformed_target (ndarray): The (mxn) ndarray. m is the number of samples,
                n is the number of feature dimensions.
        """
        return transformed_target

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def as_dict(self):
        """
        Serialize the instance into dictionary.

        Returns: dict.
        """
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "params": {},
        }

    @classmethod
    def from_dict(cls, d):
        """
        Deserialize from a dictionary
        Args:
            d: Dict, dictionary contain class initialization parameters.

        Returns: DummyScaler.

        """
        return cls()



"""Utilities package."""

from __future__ import annotations

from ._data_conversion import convert_docs, pool_from, to_array
from ._data_split import DataSplitter, ShuffleSplitter
from ._dummy import feature_dim_from_test_system, get_describer_dummy_obj
from ._inspect import get_full_args
from ._jit import njit
from ._lammps import (
    check_structures_forces_stresses,
    get_lammps_lattice_and_rotation,
    stress_format_change,
    stress_list_to_matrix,
    stress_matrix_to_list,
    write_data_from_structure,
)
from ._material import to_composition
from ._preprocessing import DummyScaler, Scaler, StandardScaler
from ._signal_processing import fft_magnitude, get_sp_method, spectrogram, wvd
from ._stats import STATS_KWARGS, Stats, get_full_stats_and_funcs, stats_list_conversion
from ._tempfile import MultiScratchDir
from ._value_profile import ConstantValue, LinearProfile, ValueProfile

__all__ = [
    "STATS_KWARGS",
    "ConstantValue",
    "DataSplitter",
    "DummyScaler",
    "LinearProfile",
    "MultiScratchDir",
    "Scaler",
    "ShuffleSplitter",
    "StandardScaler",
    "Stats",
    "ValueProfile",
    "check_structures_forces_stresses",
    "convert_docs",
    "cwt",
    "feature_dim_from_test_system",
    "fft_magnitude",
    "get_describer_dummy_obj",
    "get_full_args",
    "get_full_stats_and_funcs",
    "get_lammps_lattice_and_rotation",
    "get_sp_method",
    "njit",
    "pool_from",
    "spectrogram",
    "stats_list_conversion",
    "stress_format_change",
    "stress_list_to_matrix",
    "stress_matrix_to_list",
    "to_array",
    "to_composition",
    "write_data_from_structure",
    "wvd",
]



"""Module implements the new candidate proposal."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.linalg import solve_triangular
from scipy.optimize import minimize
from scipy.special import erfc

if TYPE_CHECKING:
    from numpy.random import RandomState
    from sklearn.gaussian_process import GaussianProcessRegressor

EPS = np.finfo(float).eps


def _trunc(values: np.ndarray, decimals: int = 3):
    """
    Truncate values to decimal places
    Args:
        values (np.ndarray): input array
        decimals (int): number of decimals to keep.

    Returns: truncated array

    """
    return np.trunc(values * 10**decimals) / 10**decimals


def ensure_rng(seed: int | None = None) -> RandomState:
    """
    Create a random number generator based on an optional seed.
    This can be an integer for a seeded rng or None for an unseeded rng.
    """
    return np.random.RandomState(seed=seed)  # pylint: disable=E1101


def predict_mean_std(x: list | np.ndarray, gpr: GaussianProcessRegressor, noise: float) -> tuple:
    """
    Speed up the gpr.predict method by manually computing the kernel operations.

    Args:
        x (list/ndarray): Query point need to be evaluated.
        gpr (GaussianProcessRegressor): A Gaussian process regressor fitted to
            known data.
        noise (float): Noise added to standard deviation if test target
            instead of GP posterior is sampled. 0 otherwise.
    """
    # cache result of K_inv computation.
    X = np.array(x)

    # todo: @yunxing the cache mechanism is causing inconsistent shapes for
    # the dot product line. Please double check. I am disabling the cache for now
    if True:  # getattr(gpr, "_K_inv", None) is None:
        L_inv = solve_triangular(gpr.L_.T, np.eye(gpr.L_.shape[0]))
        gpr._K_inv = L_inv.dot(L_inv.T)

    K_trans = gpr.kernel_(X, gpr.X_train_)
    y_mean = K_trans.dot(gpr.alpha_) + gpr._y_train_mean
    y_var = gpr.kernel_.diag(X)
    y_var = y_var - np.einsum("ij,ij->i", np.dot(K_trans, gpr._K_inv), K_trans)
    y_var += noise**2
    return (y_mean, np.sqrt(y_var))


def lhs_sample(n_intervals: int, bounds: np.ndarray, random_state: RandomState) -> np.ndarray:
    """
    Latin hypercube sampling.

    Args:
        n_intervals (int): Number of intervals.
        bounds (nd.array): Bounds for each dimension.
        random_state (RandomState): Random state.
    """
    bounds = np.array(bounds)
    dim = len(bounds)
    linspace = np.linspace(bounds[:, 0], bounds[:, 1], n_intervals + 1)
    linspace = _trunc(linspace, decimals=3)
    lower = linspace[:n_intervals, :]
    upper = linspace[1 : n_intervals + 1, :]
    _center = (lower + upper) / 2
    return np.stack([random_state.permutation(_center[:, i]) for i in range(dim)]).T


def propose_query_point(
    acquisition, scaler, gpr, y_max, noise, bounds, random_state, sampler, n_warmup=10000
) -> np.ndarray:
    """
    Strategy used to find the maximum of the acquisition function.
    It uses a combination of random sampling (cheap) and the 'L-BFGS-B'
    optimization method by first sampling `n_warmup` points at random
    and running L-BFGS-B from `n_iter` random starting points.

    Args:
        acquisition: The acquisition function.
        scaler: The Scaler used to transform features.
        gpr (GaussianProcessRegressor): A Gaussian process regressor fitted to
            known data.
        y_max (float): The current maximum target value.
        noise (float): The noise added to the acquisition function if noisy-based
            Bayesian optimization was performed.
        bounds (ndarray): The bounds of candidate points.
        random_state (RandomState): Random number generator.
        sampler (str): Sampler generating warmup points. "uniform" or "lhs".
        n_warmup (int): Number of randomly sampled points to select the initial
            point for minimization.
    """
    dim = bounds.shape[0]
    if sampler == "lhs":
        x_warmup = lhs_sample(n_warmup, bounds, random_state)
    else:
        x_warmup = _trunc(random_state.uniform(bounds[:, 0], bounds[:, 1], size=(n_warmup, dim)), decimals=3)
    x_warmup = scaler.transform(x_warmup)
    acq_warmup = acquisition(x_warmup, gpr=gpr, y_max=y_max, noise=noise)
    x_max = x_warmup[acq_warmup.argmax()]
    acq_max = acq_warmup.max()

    def min_obj(x):
        # Minimize objective is the negative acquisition function
        return -acquisition(x.reshape(-1, dim), gpr=gpr, y_max=y_max, noise=noise)

    # make sure that the initial conditions fall into the bounds
    x0 = np.clip(x_max, bounds[:, 0] + 3 * EPS, bounds[:, 1] - 3 * EPS)

    res = minimize(min_obj, x0=x0, bounds=bounds, method="L-BFGS-B")
    if -float(res.fun) >= acq_max:
        x_max = res.x
    return _trunc(scaler.inverse_transform(x_max), decimals=3)


class AcquisitionFunction:
    """An object to compute the acquisition functions."""

    def __init__(self, acq_type: str, kappa: float, xi: float):
        """
        Args:
            acq_type (str): The type of acquisition function. Three choices are given,
                ucb: Upper confidence bound,
                ei: Expected improvement,
                poi: Probability of improvement.
            kappa (float): Tradeoff parameter used in upper confidence bound formulism.
            xi (float): Tradeoff parameter between exploitation and exploration.
        """
        self.kappa = kappa
        self.xi = xi

        if acq_type not in ["ucb", "ei", "poi", "gp-ucb"]:
            err_msg = f"The utility function {acq_type} has not been implemented, please choose one of ucb, ei, or poi."
            raise NotImplementedError(err_msg)
        self.acq_type = acq_type

    def calculate(self, x: list | np.ndarray, gpr: GaussianProcessRegressor, y_max: float, noise: float) -> np.ndarray:
        """
        Calculate the value of acquisition function.

        Args:
            x (ndarray): Query point need to be evaluated.
            gpr (GaussianProcessRegressor): A Gaussian process regressor fitted to
                known data.
            y_max (float): The current maximum target value.
            noise (float): Noise added to acquisition function if noisy-based Bayesian
                optimization is performed, 0 otherwise.
        """
        if self.acq_type not in ["ucb", "ei", "poi", "gp-ucb"]:
            raise ValueError("acq type not recognized")
        if self.acq_type == "ucb":
            return self._ucb(x, gpr, self.kappa, noise)
        if self.acq_type == "ei":
            return self._ei(x, gpr, y_max, self.xi, noise)
        if self.acq_type == "poi":
            return self._poi(x, gpr, y_max, self.xi, noise)
        return self._gpucb(x, gpr, noise)

    @staticmethod
    def _ucb(x: list | np.ndarray, gpr: GaussianProcessRegressor, kappa: float, noise: float) -> np.ndarray:
        mean, std = predict_mean_std(x, gpr, noise)
        return mean + kappa * std

    @staticmethod
    def _ei(x: list | np.ndarray, gpr: GaussianProcessRegressor, y_max: float, xi: float, noise: float) -> np.ndarray:
        mean, std = predict_mean_std(x, gpr, noise)

        imp = mean - y_max - xi
        z = imp / std
        pdf = np.exp(-0.5 * z**2) / np.sqrt(2 * np.pi)
        cdf = 0.5 * erfc(-z / np.sqrt(2))
        return imp * cdf + std * pdf

    @staticmethod
    def _poi(x: list | np.ndarray, gpr: GaussianProcessRegressor, y_max: float, xi: float, noise: float) -> np.ndarray:
        mean, std = predict_mean_std(x, gpr, noise)

        z = (mean - y_max - xi) / std
        return 0.5 * erfc(-z / np.sqrt(2))

    @staticmethod
    def _gpucb(x: list | np.ndarray, gpr: GaussianProcessRegressor, noise: float) -> np.ndarray:
        if not hasattr(gpr, "X_train_"):
            raise AttributeError("GP-UCB acquisition function can not be applued.")
        T = gpr.X_train_.shape[0]
        D = gpr.X_train_.shape[1]
        kappa = np.sqrt(4 * (D + 1) * np.log(T))

        mean, std = predict_mean_std(x, gpr, noise)
        return mean + kappa * std



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This module provides NNP interatomic potential class."""
from __future__ import annotations

import itertools
import os
import re
import subprocess
from collections import OrderedDict
from shutil import which

import numpy as np
import pandas as pd
from monty.io import zopen
from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.core import Element, Lattice, Structure, units

from maml.utils import check_structures_forces_stresses, convert_docs, pool_from

from ._lammps import LammpsPotential

module_dir = os.path.dirname(__file__)
NNinput_params = loadfn(os.path.join(module_dir, "params", "NNinput.json"))


class NNPotential(LammpsPotential):
    """This class implements Neural Network Potential."""

    bohr_to_angstrom = units.bohr_to_angstrom
    eV_to_Ha = units.eV_to_Ha
    pair_style = (
        'pair_style        nnp dir "./" showew no showewsum 0 '
        "maxew 10000000 resetew yes cflength 1.8897261328 cfenergy 0.0367493254"
    )
    pair_coeff = "pair_coeff        * * {}"

    def __init__(self, name=None, param=None, weight_param=None, scaling_param=None):
        """

        Args:
            name (str): Name of force field.
            param (str): Input parameters.
            weight_param (dict): Atomic neural networks weights.
            scaling_param (dict): Scaling parameters for symmetry functions.
        """
        self.name = name if name else "NNPotential"
        self.elements = None
        self.weights = {}
        self.bs = {}
        self.atom_energy = None
        self.normalized_nodes = None
        self.epochs = None
        self.param = param if param else {}
        self.weight_param = weight_param if weight_param else {}
        self.scaling_param = scaling_param if scaling_param else None
        self.train_energy_rmse = None
        self.validation_energy_rmse = None
        self.train_forces_rmse = None
        self.validation_forces_rmse = None
        self.fitted = False

    def _line_up(self, structure, energy, forces, virial_stress):
        """
        Convert input structure, energy, forces, virial_stress to
        proper configuration format for n2p2 usage. Note that
        n2p2 takes bohr as length unit and Hartree as energy unit.

        Args:
            structure (Structure): Pymatgen Structure object.
            energy (float): DFT-calculated energy of the system.
            forces (list): The forces should have dimension
                (num_atoms, 3).
            virial_stress (list): stress should has 6 distinct
                elements arranged in order [xx, yy, zz, xy, yz, xz].
        """
        inputs = OrderedDict(
            Size=structure.num_sites,
            SuperCell=structure.lattice,
            AtomData=(structure, forces),
            Energy=energy,
            Stress=virial_stress,
        )

        lines = ["begin"]

        if "SuperCell" in inputs:
            bohr_matrix = inputs["SuperCell"].matrix / self.bohr_to_angstrom
            for vec in bohr_matrix:
                lines.append(f"lattice {vec[0]:>15.6f}{vec[1]:>15.6f}{vec[2]:>15.6f}")
        if "AtomData" in inputs:
            format_float = "atom{:>16.9f}{:>16.9f}{:>16.9f}{:>4s}{:>15.9f}{:>15.9f}{:>15.9f}{:>15.9f}{:>15.9f}"
            for _i, (site, force) in enumerate(zip(structure, forces, strict=False)):
                lines.append(
                    format_float.format(
                        *site.coords / self.bohr_to_angstrom,
                        site.species_string,
                        0.0,
                        0.0,
                        *np.array(force) * self.eV_to_Ha * self.bohr_to_angstrom,
                    )
                )
        if "Energy" in inputs:
            lines.append(f"energy  {energy * self.eV_to_Ha:f}")

        lines.append(f"charge  {structure.charge:f}")
        lines.append("end")

        return "\n".join(lines)

    def write_cfgs(self, filename, cfg_pool):
        """
        Write the formatted configuration file.

        Args:
            filename (str): The filename to be written.
            cfg_pool (list): The configuration pool contains
                structure and energy/forces properties.
        """
        lines = []
        elements = []
        for dataset in cfg_pool:
            if isinstance(dataset["structure"], dict):
                structure = Structure.from_dict(dataset["structure"])
            else:
                structure = dataset["structure"]
            energy = dataset["outputs"]["energy"]
            forces = dataset["outputs"]["forces"]
            virial_stress = dataset["outputs"]["virial_stress"]

            lines.append(self._line_up(structure, energy, forces, virial_stress))

            elements.extend(structure.species)

        if self.elements is None:
            self.elements = [element.name for element in sorted(set(elements))]

        with open(filename, "w") as f:
            f.write("\n".join(lines))

        return filename

    def write_input(self, **kwargs):
        """
        Write input.nn file to train the Neural Network Potential.

        Args:
            atom_energy (float): Atomic reference energy.

            kwargs:
                General nnp settings:
                    atom_energy (dict): Free atom reference energy for each specie.
                    cutoff_type (int): Type of cutoff function. Default to 1
                        (i.e., cosine function).
                    scale_features (int): Determine the method to scale the
                        symmetry function.
                        0: no scaling.
                        1: scale_symmetry_functions.
                        2: center_symmetry_functions.
                        3. scale_symmetry_functions_sigma.
                    scale_min_short (float): Minimum value for scaling.
                        Default to 0.0.
                    scale_max_short (float): Maximum value for scaling.
                        Default to 1.
                    hidden_layers (list): List of the numbers of
                        nodes in each hidden layer.
                    activations (str): Activation function for each hidden layer.
                        't': tanh, 's': logistic, 'p': softplus.
                    normalize_nodes (boolean): Whether to normalize input of nodes.

                Additional settings for training:
                    epoch (int): Number of training epochs.
                    updater_type (int): Weight update method
                        0: gradient Descent, 1: Kalman filter.
                    parallel_mode (int): Training parallelization used.
                        Default to serial mode.
                    jacobian_mode (int): Jacobian computation mode.
                        0: Summation to single gradient,
                        1: Per-task summed gradient,
                        2: Full Jacobian.
                    update_strategy (int): Update strategy.
                        0: combined, 1: per-element.
                    selection_mode (int): Update candidate selection mode.
                        0: random, 1: sort, 2: threshold
                    task_batch_size_energy (int): Number of energy update
                        candidates prepared per task for each update.
                    task_batch_size_force (int): Number of force update
                        candidates prepared per task for each update.
                    test_fraction (float): Fraction of structures kept for
                        testing.
                    force_weight (float): Weight of force updates relative
                        to energy updates. Default to 10.0
                    short_energy_fraction (float): Fraction of energy updates
                        per epoch. Default to 1.0.
                    short_force_fraction (float): Fraction of force updates
                        per epoch. Default to 0.02315.
                    short_energy_error_threshold (float): RMSE threshold for
                        energy update candidates. Default to 0.0.
                    short_force_error_threshold (float): RMSE threshold for
                        force update candidates. Default to 1.0.
                    rmse_threshold_trials (int): Maximum number of RMSE
                        threshold trials. Default to 3.
                    weights_min (float): Minimum value for initial random
                        weights. Default to -1.
                    weights_max (float): Maximum value for initial random
                        weights. Default to 1.
                    write_trainpoints (int): Write energy comparison every
                        this many epochs. Default to 1.
                    write_trainforces (int): Write force comparison every
                        this many epochs. Default to 1.
                    write_weights_epoch (int): Write weights every this many
                        epochs. Default to 1.
                    write_neuronstats (int): Write neuron statistics every
                        this many epochs. Default to 1.

                    # Kalman Filter
                    kalman_type (int): Kalman filter type. Default to 0.
                    kalman_epsilon (float): General Kalman filter parameter
                        epsilon. Default to 0.01.
                    kalman_q0 (float): General Kalman filter parameter q0.
                        Default to 0.01.
                    kalman_qtau (float): General Kalman filter parameter
                        qtau. Default to 2.302.
                    kalman_qmin (float): General Kalman filter parameter qmin.
                        Default to 1e-6.
                    kalman_eta (float): Standard Kalman filter parameter eta.
                        Default to 0.01.
                    kalman_etatau (float): Standard Kalman filter parameter
                        etatau. Defaults to 2.302.
                    kalman_etamax (float): Standard Kalman filter parameter
                        etamax. Default to 1.0.

                Symmetry functions:
                    r_cut (float): Cutoff distance (unit: Å).
                    r_etas (numpy.array): η in radial function.
                    r_shift (numpy.array): Rs in radial function.
                    a_etas (numpy.array): η in angular function.
                    zetas (numpy.array): ζ in angular function.
                    lambdas (numpy.array): λ in angular function. Default to (1, -1).
        """
        filename = "input.nn"

        head_formatter = "{:<32s}{value}"
        type2_format = "symfunction_short {central_atom}  2 {neighbor_atom}    {r_eta:.7f}    {rs:.7f}    {rcut:.7f}"
        type3_format = (
            "symfunction_short {central_atom}  3 {neighbor_atom1} "
            "{neighbor_atom2}    {a_eta:.7f} {lambd:>2d} {zeta:.7f}   "
            "{rcut:.7f}"
        )

        lines = [
            head_formatter.format("number_of_elements", value=len(self.elements)),
            head_formatter.format("elements", value=" ".join(self.elements)),
        ]

        PARAMS = {
            "general": ["cutoff_type", "scale_features", "scale_min_short", "scale_max_short", "hidden_layers"],
            "additional": [
                "epochs",
                "updater_type",
                "parallel_mode",
                "jacobian_mode",
                "update_strategy",
                "selection_mode",
                "task_batch_size_energy",
                "task_batch_size_force",
                "random_seed",
                "test_fraction",
                "force_weight",
                "short_energy_fraction",
                "short_force_fraction",
                "short_energy_error_threshold",
                "short_force_error_threshold",
                "rmse_threshold_trials",
                "weights_min",
                "weights_max",
                "write_trainpoints",
                "write_trainforces",
                "write_weights_epoch",
                "write_neuronstats",
                "kalman_type",
                "kalman_epsilon",
                "kalman_q0",
                "kalman_qtau",
                "kalman_qmin",
                "kalman_eta",
                "kalman_etatau",
                "kalman_etamax",
            ],
        }
        if self.fitted:
            if self.param.get("atom_energy"):
                for specie in self.elements:
                    lines.append(
                        head_formatter.format(
                            "atom_energy", value=" ".join([specie, str(self.param.get("atom_energy").get(specie))])
                        )
                    )
            for tag in PARAMS.get("general"):
                if tag == "scale_features":
                    lines.append(NNinput_params.get("general").get(tag).get(self.param.get(tag)))
                elif tag == "hidden_layers":
                    layers = self.param.get("hidden_layers")
                    activations = self.param.get("activations")
                    lines.append(head_formatter.format("global_hidden_layers_short", value=len(layers)))
                    lines.append(head_formatter.format("global_nodes_short", value=" ".join([str(i) for i in layers])))
                    lines.append(
                        head_formatter.format(
                            "global_activation_short", value=" ".join([activations] * len(layers) + ["l"])
                        )
                    )
                else:
                    lines.append(head_formatter.format(tag, value=self.param.get(tag)))
            if self.normalized_nodes:
                lines.append("normalize_nodes")

            for tag in PARAMS.get("additional"):
                lines.append(head_formatter.format(tag, value=self.param.get(tag)))
            lines.append("use_short_forces")

            r_cut = self.param.get("r_cut")
            r_cut /= self.bohr_to_angstrom
            r_shift = np.array(self.param.get("r_shift"))
            r_shift /= self.bohr_to_angstrom

            for central_atom in self.elements:
                for neighbor_atom1 in self.elements:
                    for r_eta, rs in itertools.product(self.param.get("r_etas"), r_shift):
                        lines.append(
                            type2_format.format(
                                central_atom=central_atom, neighbor_atom=neighbor_atom1, r_eta=r_eta, rs=rs, rcut=r_cut
                            )
                        )
            for central_atom in self.elements:
                for neighbor_atom1, neighbor_atom2 in itertools.combinations_with_replacement(self.elements, 2):
                    for a_eta, lambd, zeta in itertools.product(
                        self.param.get("a_etas"), self.param.get("lambdas"), self.param.get("zetas")
                    ):
                        lines.append(
                            type3_format.format(
                                central_atom=central_atom,
                                neighbor_atom1=neighbor_atom1,
                                neighbor_atom2=neighbor_atom2,
                                a_eta=a_eta,
                                lambd=lambd,
                                zeta=zeta,
                                rcut=r_cut,
                            )
                        )
        else:
            if kwargs.get("atom_energy"):
                for specie in self.elements:
                    lines.append(
                        head_formatter.format(
                            "atom_energy", value=" ".join([specie, str(kwargs.get("atom_energy").get(specie))])
                        )
                    )
                self.param.update({"atom_energy": kwargs.get("atom_energy")})
            for tag in PARAMS.get("general"):
                if tag == "scale_features":
                    value = kwargs.get(tag) if kwargs.get(tag) is not None else "1"
                    lines.append(NNinput_params.get("general").get(tag).get(value))
                    self.param.update({tag: value})
                elif tag == "hidden_layers":
                    layers = kwargs.get(tag, NNinput_params.get("general").get(tag))
                    self.param.update({tag: layers})
                    activations = kwargs.get("activations", NNinput_params.get("general").get("activations"))
                    self.param.update({"activations": activations})
                    lines.append(head_formatter.format("global_hidden_layers_short", value=len(layers)))
                    lines.append(head_formatter.format("global_nodes_short", value=" ".join([str(i) for i in layers])))
                    lines.append(
                        head_formatter.format(
                            "global_activation_short", value=" ".join([activations] * len(layers) + ["l"])
                        )
                    )
                else:
                    value = kwargs.get(tag, NNinput_params.get("general").get(tag))
                    lines.append(head_formatter.format(tag, value=value))
                    self.param.update({tag: value})
            if kwargs.get("normalize_nodes"):
                lines.append("normalize_nodes")
                self.param.update({"normalize_nodes": True})

            for tag in PARAMS.get("additional"):
                value = kwargs.get(tag, NNinput_params.get("additional").get(tag))
                lines.append(head_formatter.format(tag, value=value))
                self.param.update({tag: value})
            lines.append("use_short_forces")

            r_cut = (
                kwargs.get("r_cut")
                if kwargs.get("r_cut") is not None
                else NNinput_params.get("symmetry_function").get("r_cut")
            )
            self.param.update({"r_cut": r_cut})
            r_cut /= self.bohr_to_angstrom
            r_etas = (
                kwargs.get("r_etas")
                if kwargs.get("r_etas") is not None
                else NNinput_params.get("symmetry_function").get("r_etas")
            )
            self.param.update({"r_etas": r_etas})
            r_shift = (
                kwargs.get("r_shift")
                if kwargs.get("r_shift") is not None
                else NNinput_params.get("symmetry_function").get("r_shift")
            )
            self.param.update({"r_shift": r_shift})
            r_shift = np.array(r_shift)
            r_shift /= self.bohr_to_angstrom
            a_etas = (
                kwargs.get("a_etas")
                if kwargs.get("a_etas") is not None
                else NNinput_params.get("symmetry_function").get("a_etas")
            )
            self.param.update({"a_etas": a_etas})
            zetas = (
                kwargs.get("zetas")
                if kwargs.get("zetas") is not None
                else NNinput_params.get("symmetry_function").get("zetas")
            )
            self.param.update({"zetas": zetas})
            lambdas = (
                kwargs.get("lambdas")
                if kwargs.get("lambdas") is not None
                else NNinput_params.get("symmetry_function").get("lambdas")
            )
            self.param.update({"lambdas": lambdas})

            for central_atom in self.elements:
                for neighbor_atom1 in self.elements:
                    for r_eta, rs in itertools.product(self.param.get("r_etas"), r_shift):
                        lines.append(
                            type2_format.format(
                                central_atom=central_atom, neighbor_atom=neighbor_atom1, r_eta=r_eta, rs=rs, rcut=r_cut
                            )
                        )
            for central_atom in self.elements:
                for neighbor_atom1, neighbor_atom2 in itertools.combinations_with_replacement(self.elements, 2):
                    for a_eta, lambd, zeta in itertools.product(
                        self.param.get("a_etas"), self.param.get("lambdas"), self.param.get("zetas")
                    ):
                        lines.append(
                            type3_format.format(
                                central_atom=central_atom,
                                neighbor_atom1=neighbor_atom1,
                                neighbor_atom2=neighbor_atom2,
                                a_eta=a_eta,
                                lambd=lambd,
                                zeta=zeta,
                                rcut=r_cut,
                            )
                        )

            self.num_symm_functions = sum(len(list(itertools.product(r_etas, r_shift))) for _ in self.elements) + sum(
                len(list(itertools.product(a_etas, lambdas, zetas)))
                for _, _ in itertools.combinations_with_replacement(self.elements, 2)
            )

            self.layer_sizes = [self.num_symm_functions, *self.param.get("hidden_layers")]

        with open(filename, "w") as f:
            f.write("\n".join(lines))

        self.fitted = True

        return filename

    def load_input(self, filename="input.nn"):
        """
        Load input file from trained Neural Network Potential.

        Args:
            filename (str): The input filename.
        """
        PARAMS = {
            "general": ["cutoff_type", "scale_features", "scale_min_short", "scale_max_short", "hidden_layers"],
            "additional": [
                "epochs",
                "updater_type",
                "parallel_mode",
                "jacobian_mode",
                "update_strategy",
                "selection_mode",
                "task_batch_size_energy",
                "task_batch_size_force",
                "random_seed",
                "test_fraction",
                "force_weight",
                "short_energy_fraction",
                "short_force_fraction",
                "short_energy_error_threshold",
                "short_force_error_threshold",
                "rmse_threshold_trials",
                "weights_min",
                "weights_max",
                "write_trainpoints",
                "write_trainforces",
                "write_weights_epoch",
                "write_neuronstats",
                "kalman_type",
                "kalman_epsilon",
                "kalman_q0",
                "kalman_qtau",
                "kalman_qmin",
                "kalman_eta",
                "kalman_etatau",
                "kalman_etamax",
            ],
        }

        def str_formatify(string):
            return float(string) if "." in string or "e" in string else int(string)

        param = {}
        with open(filename) as f:
            lines = f.readlines()
        df = pd.DataFrame([line.split() for line in lines if "#" not in line])
        self.elements = sorted(
            (element for element in np.ravel(df[df[0] == "elements"])[1:] if element is not None),
            key=lambda x: Element(x),
        )

        atom_energy = {}
        for atom, energy in zip(
            np.array(df[df[0] == "atom_energy"])[:, 1], np.array(df[df[0] == "atom_energy"])[:, 2], strict=False
        ):
            atom_energy[atom] = float(energy)
        param.update({"atom_energy": atom_energy})
        for tag in PARAMS.get("general"):
            if tag == "scale_features":
                scale_features = "1" if len(df[df[0] == "scale_symmetry_functions"]) != 0 else 0
                param.update({"scale_features": scale_features})
            elif tag == "hidden_layers":
                hidden_layers = [int(neuron) for neuron in np.array(df[df[0] == "global_nodes_short"])[0][1:] if neuron]
                param.update({"hidden_layers": hidden_layers})
                activations = np.array(df[df[0] == "global_activation_short"])[0][1]
                param.update({"activations": activations})
            else:
                value = str_formatify(np.array(df[df[0] == tag])[0][1])
                param.update({tag: value})
        if len(df[df[0] == "normalize_nodes"]) != 0:
            param.update({"normalize_nodes": True})

        for tag in PARAMS.get("additional"):
            value = str_formatify(np.array(df[df[0] == tag])[0][1])
            param.update({tag: value})

        r_cut = np.sort(np.array(df[(df[0] == "symfunction_short") & (df[2] == "2")][6], dtype=np.float64))[0]
        r_cut = float(f"{r_cut * units.bohr_to_angstrom:.1f}")
        param.update({"r_cut": r_cut})
        r_etas = np.sort(
            np.array(np.unique(df[(df[0] == "symfunction_short") & (df[2] == "2")][4]), dtype=np.float64)
        ).tolist()
        param.update({"r_etas": r_etas})
        r_shift = np.sort(
            np.array(np.unique(df[(df[0] == "symfunction_short") & (df[2] == "2")][5]), dtype=np.float64)
        ).tolist()
        r_shift = [float(f"{r * units.bohr_to_angstrom:.1f}") for r in r_shift]
        param.update({"r_shift": r_shift})
        a_etas = np.sort(
            np.array(np.unique(df[(df[0] == "symfunction_short") & (df[2] == "3")][5]), dtype=np.float64)
        ).tolist()
        param.update({"a_etas": a_etas})
        lambdas = np.sort(
            np.array(np.unique(df[(df[0] == "symfunction_short") & (df[2] == "3")][6]), dtype=np.int64)
        ).tolist()
        param.update({"lambdas": lambdas})
        zetas = np.sort(
            np.array(np.unique(df[(df[0] == "symfunction_short") & (df[2] == "3")][7]), dtype=np.float64)
        ).tolist()
        param.update({"zetas": zetas})
        self.num_symm_functions = sum(len(list(itertools.product(r_etas, r_shift))) for _ in self.elements) + sum(
            len(list(itertools.product(a_etas, lambdas, zetas)))
            for _, _ in itertools.combinations_with_replacement(self.elements, 2)
        )
        self.layer_sizes = [self.num_symm_functions, *hidden_layers]
        self.param = param

    def load_weights(self, weights_filename, specie):
        """
        Load weights file of trained Neural Network Potential.

        Args:
            weights_filename (str): The weights file.
            specie (str): The name of specie.
        """
        if not self.weights.get(specie) or not self.bs.get(specie) or not self.weight_param:
            self.weights[specie] = []
            self.bs[specie] = []
            self.weight_param[specie] = []

        with open(weights_filename) as f:
            weights_lines = f.readlines()

        weight_param = pd.DataFrame([line.split() for line in weights_lines if "#" not in line])
        weight_param.columns = ["value", "type", "index", "start_layer", "start_neuron", "end_layer", "end_neuron"]

        for layer_index in range(1, len(self.layer_sizes)):
            weights_group = weight_param[
                (weight_param["start_layer"] == str(layer_index - 1)) & (weight_param["end_layer"] == str(layer_index))
            ]

            weights = np.reshape(
                np.array(weights_group["value"], dtype=np.float64),
                (self.layer_sizes[layer_index - 1], self.layer_sizes[layer_index]),
            )
            self.weights[specie].append(weights)

            bs_group = weight_param[(weight_param["type"] == "b") & (weight_param["start_layer"] == str(layer_index))]
            bs = np.array(bs_group["value"], dtype=np.float64)
            self.bs[specie].append(bs)

        self.weight_param[specie] = weight_param

    def load_scaler(self, scaling_filename):
        """
        Load scaling info of trained Neural Network Potential.

        Args:
            scaling_filename (str): The scaling file.
        """
        with open(scaling_filename) as f:
            scaling_lines = f.readlines()
        scaling_param = pd.DataFrame([line.split() for line in scaling_lines if "#" not in line])
        self.scaling_param = scaling_param

    def read_cfgs(self, filename="output.data"):
        """
        Read the configuration file.

        Args:
            filename (str): The configuration file to be read.
        """
        data_pool = []
        with zopen(filename, "rt") as f:
            lines = f.read()

        block_pattern = re.compile("begin\n(.*?)end", re.S)
        lattice_pattern = re.compile("lattice(.*?)\n")
        position_pattern = re.compile("atom(.*?)\n")
        energy_pattern = re.compile("energy(.*?)\n")

        for block in block_pattern.findall(lines):
            d = {"outputs": {}}
            lattice_str = lattice_pattern.findall(block)
            lattice = Lattice(
                np.array([latt.split() for latt in lattice_str], dtype=np.float64) * self.bohr_to_angstrom
            )
            position_str = position_pattern.findall(block)
            positions = pd.DataFrame([pos.split() for pos in position_str])
            positions.columns = ["x", "y", "z", "specie", "charge", "atomic_energy", "fx", "fy", "fz"]
            coords = np.array(positions.loc[:, ["x", "y", "z"]], dtype=np.float64)
            coords = coords * self.bohr_to_angstrom
            species = np.array(positions["specie"])
            forces = np.array(positions.loc[:, ["fx", "fy", "fz"]], dtype=np.float64)
            forces = forces / self.eV_to_Ha / self.bohr_to_angstrom
            energy_str = energy_pattern.findall(block)[0]
            energy = float(energy_str.lstrip()) / self.eV_to_Ha
            struct = Structure(lattice=lattice, species=species, coords=coords, coords_are_cartesian=True)
            d["structure"] = struct.as_dict()
            d["outputs"]["energy"] = energy
            d["outputs"]["forces"] = forces
            d["num_atoms"] = len(struct)

            data_pool.append(d)
        _, df = convert_docs(docs=data_pool)
        return data_pool, df

    def write_param(self):
        """Write optimized weights file to perform energy and force prediction."""
        if self.weight_param is None or self.scaling_param is None:
            raise RuntimeError("The parameters should be provided.")
        for specie in self.elements:
            weights_filename = ".".join(["weights", str(Element(specie).number).zfill(3), "data"])
            weight_formatter = "{:>18s}{:>2s}{:>10s}{:>6s}{:>6s}{:>6s}{:>6s}"
            bias_formatter = "{:>18s}{:>2s}{:>10s}{:>6s}{:>6}"
            lines = []
            for i in range(self.weight_param[specie].shape[0]):
                if self.weight_param[specie].iloc[i]["type"] == "a":
                    lines.append(weight_formatter.format(*self.weight_param[specie].iloc[i]))
                else:
                    lines.append(bias_formatter.format(*self.weight_param[specie].iloc[i]))

            with open(weights_filename, "w") as f:
                f.writelines("\n".join(lines))

        scaling_filename = "scaling.data"
        scaling_formatter = "{:>4s}{:>5s}  {:>22s} {:>22s} {:>22s} {:.>22s}"
        scaling_lines = []
        for i in range(self.num_symm_functions * len(self.elements)):
            scaling_lines.append(scaling_formatter.format(*self.scaling_param.iloc[i]))
        with open(scaling_filename, "w") as f:
            f.writelines("\n".join(scaling_lines))

        self.write_input()

        return [self.pair_style, self.pair_coeff.format(self.param.get("r_cut") + 1e-2)]

    def train(self, train_structures, train_energies, train_forces, train_stresses=None, **kwargs):
        """
        Training data with moment tensor method.

        Args:
            train_structures ([Structure]): The list of Pymatgen Structure object.
                energies ([float]): The list of total energies of each structure
                in structures list.
            train_energies ([float]): List of total energies of each structure in
                structures list.
            train_forces ([np.array]): List of (m, 3) forces array of each structure
                with m atoms in structures list. m can be varied with each
                single structure case.
            train_stresses (list): List of (6, ) virial stresses of each
                structure in structures list.
            kwargs: Parameters in write_input method.
        """
        if not which("nnp-train"):
            raise RuntimeError("NNP Trainer has not been found.")
        train_structures, train_forces, train_stresses = check_structures_forces_stresses(
            train_structures, train_forces, train_stresses
        )
        train_pool = pool_from(train_structures, train_energies, train_forces, train_stresses)
        atoms_filename = "input.data"

        with ScratchDir("."):
            _ = self.write_cfgs(filename=atoms_filename, cfg_pool=train_pool)
            output = "training_output"

            self.write_input(**kwargs)
            with subprocess.Popen(["nnp-scaling", "100"], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p_scaling:
                stdout, stderr = p_scaling.communicate()
                rc = p_scaling.returncode
            if rc != 0:
                error_msg = f"n2p2 exited with return code {rc}"
                msg = stderr.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += ", "
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            with open(output, "w") as f, subprocess.Popen(["nnp-train"], stdout=f, stderr=subprocess.PIPE) as p_train:
                stdout, stderr = p_train.communicate()
                rc = p_train.returncode

            if rc != 0:
                error_msg = f"n2p2 exited with return code {rc}"
                msg = stderr.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += ", "
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            with zopen(output) as f:
                error_lines = f.read()

            energy_rmse_pattern = re.compile(r"ENERGY\s*\S*\s*(\S*)\s*(\S*).*?\n")
            forces_rmse_pattern = re.compile(r"FORCE\s*\S*\s*(\S*)\s*(\S*).*?\n")
            errors = np.array(energy_rmse_pattern.findall(error_lines), dtype=np.float64).T.tolist()
            self.train_energy_rmse = errors[0]
            self.validation_energy_rmse = errors[1]

            errors = np.array(forces_rmse_pattern.findall(error_lines), dtype=np.float64).T.tolist()
            self.train_forces_rmse = errors[0]
            self.validation_forces_rmse = errors[1]

            for specie in self.elements:
                weights_filename = (
                    f"weights.{str(Element(specie).number).zfill(3)}.{str(self.param['epochs']).zfill(6)}.out"
                )
                self.weights[specie] = []
                self.bs[specie] = []
                self.weight_param[specie] = []
                self.load_weights(weights_filename, specie)
            self.load_scaler("scaling.data")

        return rc

    def evaluate(self, test_structures, test_energies, test_forces, test_stresses=None):
        """
        Evaluate energies, forces and stresses of structures with trained
        interatomic potentials.

        Args:
            test_structures ([Structure]): List of Pymatgen Structure Objects.
            test_energies ([float]): List of DFT-calculated total energies of
                each structure in structures list.
            test_forces ([np.array]): List of DFT-calculated (m, 3) forces of
                each structure with m atoms in structures list. m can be varied
                with each single structure case.
            test_stresses (list): List of DFT-calculated (6, ) viriral stresses
                of each structure in structures list.
        """
        if not which("nnp-predict"):
            raise RuntimeError("NNP Predictor has not been found.")

        original_file = "input.data"
        predict_file = "output.data"
        test_structures, test_forces, test_stresses = check_structures_forces_stresses(
            test_structures, test_forces, test_stresses
        )
        predict_pool = pool_from(test_structures, test_energies, test_forces, test_stresses)
        with ScratchDir("."):
            _, _ = self.write_param()
            original_file = self.write_cfgs(original_file, cfg_pool=predict_pool)
            _, df_orig = self.read_cfgs(original_file)

            input_filename = self.write_input()

            dfs = []
            for data in predict_pool:
                _ = self.write_cfgs(original_file, cfg_pool=[data])
                with subprocess.Popen(
                    ["nnp-predict", input_filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE
                ) as p_evaluation:
                    stdout, stderr = p_evaluation.communicate()
                    rc = p_evaluation.returncode
                if rc != 0:
                    error_msg = f"n2p2 exited with return code {rc}"
                    msg = stderr.decode("utf-8").split("\n")[:-1]
                    try:
                        error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                        error_msg += ", ".join(msg[error_line:])
                    except Exception:
                        error_msg += ", "
                        error_msg += msg[-1]
                    raise RuntimeError(error_msg)

                _, df = self.read_cfgs(predict_file)
                dfs.append(df)
            df_predict = pd.concat(dfs, ignore_index=True)

        return df_orig, df_predict

    @staticmethod
    def from_config(input_filename, scaling_filename, weights_filenames):
        """
        Initialize potentials with parameters file.

        Args:
            input_filename (str): The file storing the input configuration of
                Neural Network Potential.
            scaling_filename (str): The file storing scaling info of
                Neural Network Potential.
            weights_filenames (list): List of files storing weights of each specie in
                Neural Network Potential.
        """
        nnp = NNPotential()
        nnp.load_input(input_filename)
        nnp.load_scaler(scaling_filename)
        if len(nnp.elements) != len(weights_filenames):
            raise ValueError(f"{len(nnp.elements)} weights files should be given to " + " ".join(nnp.elements))
        for weights_filename, specie in zip(weights_filenames, nnp.elements, strict=False):
            nnp.load_weights(weights_filename, specie)
        nnp.fitted = True

        return nnp



"""MAML describers base classes."""

from __future__ import annotations

import abc
import logging
import tempfile
from typing import TYPE_CHECKING, Any

from joblib import Parallel, cpu_count, delayed
from monty.json import MSONable
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_memory
from tqdm import tqdm

from maml.utils import feature_dim_from_test_system

from ._feature_batch import get_feature_batch

if TYPE_CHECKING:
    import numpy as np

_ALLOWED_DATA = ("number", "structure", "molecule", "spectrum")

_DESCRIBER_TYPES = [
    "composition",
    "site",
    "structure",
    "general",
    "band_structure",
    "spectrum",
]

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BaseDescriber(BaseEstimator, TransformerMixin, MSONable, metaclass=abc.ABCMeta):
    """
    Base class for a Describer. A describers converts an object to a descriptor,
    typically a numerical representation useful for machine learning.
    The output for the describers can be a single DataFrame/numpy.ndarray or
    a list of DataFrame/numpy.ndarray.
    """

    def __init__(self, **kwargs):
        """
        Base estimator with the following allowed keyword args.

            memory (bool/str/joblib.Memory): The path or Memory for caching the computational
                results, default None means no cache.
            verbose (bool): Whether to show the progress of feature calculations.
            n_jobs (int): The number of parallel jobs. 0 means no parallel computations.
                If this value is set to negative or greater than the total cpu
                then n_jobs is set to the number of cpu on system.
            feature_batch (str): method to batch a list of features into one

        Args:
            **kwargs: keyword args that contain possibly memory (str/joblib.Memory),
                verbose (bool), n_jobs (int)
        """
        allowed_kwargs = ["memory", "verbose", "n_jobs", "feature_batch"]

        for k in kwargs:
            if k not in allowed_kwargs:
                raise TypeError(f"{k} not allowed as kwargs")
        memory = kwargs.get("memory")
        if isinstance(memory, bool):
            memory = tempfile.mkdtemp()
            logger.info(f"Created temporary directory {memory}")
        verbose = kwargs.get("verbose", False)
        n_jobs = kwargs.get("n_jobs", 0)

        self.memory = check_memory(memory)
        self.verbose = verbose
        # find out the number of parallel jobs
        if (n_jobs < 0) or (n_jobs > cpu_count()):
            n_jobs = cpu_count()
            logger.info(f"Using {n_jobs} jobs for computation")
        self.n_jobs = n_jobs
        self.feature_batch = get_feature_batch(kwargs.get("feature_batch"))

    def fit(self, x: Any, y: Any = None) -> BaseDescriber:
        """
        Place holder for fit API.

        Args:
            x: Any inputs
            y: Any outputs

        Returns: self

        """
        return self

    def transform_one(self, obj: Any) -> list[np.ndarray] | np.ndarray:
        """Transform an object."""
        raise NotImplementedError

    def transform(self, objs: list[Any]) -> Any:
        """
        Transform a list of objs. If the return data is DataFrame,
        use df.xs(index, level='input_index') to get the result for the i-th object.

        Args:
            objs (list): A list of objects.

        Returns:
            One or a list of pandas data frame/numpy ndarray
        """
        cached_transform_one = self.memory.cache(_transform_one)

        if self.verbose:
            objs = tqdm(objs)

        # run the featurizer
        if self.n_jobs == 0:
            features = [cached_transform_one(self, obj) for obj in objs]
        else:
            features = Parallel(n_jobs=self.n_jobs)(delayed(cached_transform_one)(self, obj) for obj in objs)

        multi_output = self._is_multi_output()
        if not multi_output:
            features = [features]
        batched_features = [self.feature_batch(i) for i in list(*zip(features, strict=False))]  # type: ignore
        return batched_features if multi_output else batched_features[0]

    def _is_multi_output(self) -> bool:
        tags = self._get_tags()
        return tags["multioutput"]  # this is from BaseEstimator

    def clear_cache(self):
        """Clear cache."""
        if self.memory.location is not None:
            self.memory.clear()

    @property
    def feature_dim(self):
        """
        Feature dimension, useful when certain models need to specify
        the feature dimension, e.g., MLP models.
        """
        return feature_dim_from_test_system(self)


def _transform_one(describer: BaseDescriber, obj: Any) -> list[np.ndarray] | np.ndarray:
    """
    A wrapper to make a pure function.

    Args:
        describer (BaseDescriber): a describers
        obj: an object (structure/molecule) to transform.

    Returns:
        np.ndarray
    """
    return describer.transform_one(obj)


class DummyDescriber(BaseDescriber):
    """Dummy Describer that does nothing."""

    def transform_one(self, obj: Any):
        """
        Does nothing but return the original features.

        Args:
            obj: Any inputs

        Returns: Any outputs.

        """
        return obj


class SequentialDescriber(Pipeline):
    """A thin wrapper of sklearn Pipeline."""

    def __init__(self, describers: list, **kwargs):
        """
        Put a list of describers into one pipeline.

        Args:
            describers (list): a list of describers that will be applied
                consecutively
            **kwargs: pass to pipeline.
        """
        steps = [(i.__class__.__name__, i) for i in describers]
        super().__init__(steps, **kwargs)


def describer_type(dtype: str):
    """
    Decorate to set describers class type.

    Args:
        dtype (str): describers type, e.g., site, composition, structure etc.

    Return:
        wrapped class
    """

    def wrapped_describer(klass):
        klass.describer_type = dtype
        return klass

    return wrapped_describer



"""Inspect function args."""

from __future__ import annotations

from inspect import signature
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable


def get_full_args(func: Callable) -> list:
    """
    Get args from function.

    Args:
        func (callable): function to determine the args
    """
    all_params = list(signature(func).parameters.keys())
    return [i for i in all_params if i != "self"]


def get_param_types(func):
    """
    Get param and type info.

    Args:
        func (callable): function to determine the arg types
    """
    all_params = get_full_args(func)
    params = signature(func).parameters
    param_dict = {}

    for i in all_params:
        annotation = params[i].annotation
        param_type = getattr(annotation, "__args__", None)
        if param_type is not None:
            param_dict[i] = param_type[0].__name__
        else:
            param_dict[i] = annotation.__name__
    return param_dict



"""
Describer for converting (structural) objects into models-readable
numeric vectors or tensors.
"""

from __future__ import annotations

from ._composition import ElementProperty, ElementStats
from ._m3gnet import M3GNetSite, M3GNetStructure
from ._matminer import wrap_matminer_describer
from ._megnet import MEGNetSite, MEGNetStructure
from ._rdf import RadialDistributionFunction
from ._site import BispectrumCoefficients, BPSymmetryFunctions, SiteElementProperty, SmoothOverlapAtomicPosition
from ._structure import (
    CoulombEigenSpectrum,
    CoulombMatrix,
    DistinctSiteProperty,
    RandomizedCoulombMatrix,
    SortedCoulombMatrix,
)

__all__ = [
    "BPSymmetryFunctions",
    "BispectrumCoefficients",
    "CoulombEigenSpectrum",
    "CoulombMatrix",
    "DistinctSiteProperty",
    "ElementProperty",
    "ElementStats",
    "M3GNetSite",
    "M3GNetStructure",
    "MEGNetSite",
    "MEGNetStructure",
    "RadialDistributionFunction",
    "RandomizedCoulombMatrix",
    "SiteElementProperty",
    "SmoothOverlapAtomicPosition",
    "SortedCoulombMatrix",
    "wrap_matminer_describer",
]



# Copyright (c) Materials Virtual Lab
# Distributed under the terms of the BSD License.

"""This package contains PotentialMixin classes representing Interatomic Potentials."""
from __future__ import annotations

from ._base import Potential, PotentialMixin
from ._gap import GAPotential
from ._lammps import (
    DefectFormation,
    ElasticConstant,
    EnergyForceStress,
    LatticeConstant,
    LMPRelaxationCalculator,
    LMPStaticCalculator,
    NudgedElasticBand,
    SpectralNeighborAnalysis,
    SurfaceEnergy,
    get_default_lmp_exe,  # ; noqa
)
from ._mtp import MTPotential
from ._nnp import NNPotential
from ._snap import SNAPotential

__all__ = [
    "DefectFormation",
    "ElasticConstant",
    "EnergyForceStress",
    "GAPotential",
    "LMPStaticCalculator",
    "LatticeConstant",
    "MTPotential",
    "NNPotential",
    "NudgedElasticBand",
    "Potential",
    "SNAPotential",
    "SpectralNeighborAnalysis",
    "get_default_lmp_exe",
]



"""Module defines feature objects for GB energy model."""

from __future__ import annotations


class my_quant:
    """An object to describe GB quantities."""

    def __init__(self, str_name: str, latex_name: str | None = None, latex_unit: str | None = None):
        """

        Args:
            str_name (str): string rep of the quant
            latex_name (): latex rep of the quant
            latex_unit (): latex rep of the quant unit.
        """
        self.str_name = str_name
        self.latex_name = latex_name
        self.latex_unit = latex_unit

    @property
    def name(self):
        """

        Returns:
            string rep of the quant.
        """
        return self.str_name

    @property
    def latex(self):
        """

        Returns:
            latex rep of the quant.
        """
        return self.latex_name

    @property
    def unit(self):
        """

        Returns:
            latex rep of the quant unit.
        """
        return self.latex_unit


initial_fnames = [
    r"$d_{GB}$",
    r"$d_{rot}$",
    r"cos($\theta$)",
    r"$E_{coh}$",
    r"$\widebar{{\Delta} BL}$",
    r"$G$",
    r"$a_{0}$",
    r"$\gamma_{GB}^{DFT}$",
]
str2lat = {
    "d_gb": r"$d_{GB}$",
    "d_rot": r"$d_{rot}$",
    "sin(theta)": r"sin($\theta$)",
    "cos(theta)": r"cos($\theta$)",
    "e_coh": r"$E_{coh}$",
    "G": r"$G$",
    "a_0": r"$a_{0}$",
    "a_r": r"$a_{r}$",
    "mean_delta_bl": r"$\bar{{\Delta} BL}$",
    "mean_bl": r"$\bar{BL}$",
    "Density": "Density",
    "e_gb": r"$\gamma_{GB}$",
}
features = ["d_gb", "d_rot", "cos(theta)", "mean_delta_bl", "G", "mean_bl"]

e_coh = my_quant("e_coh", str2lat["e_coh"], "J")
a0 = my_quant("a_0", str2lat["a_0"], r"$\AA$")
ar = my_quant("a_r", str2lat["a_r"], r"$\AA$")
cos_theta = my_quant("cos(theta)", str2lat["cos(theta)"])
sin_theta = my_quant("sin(theta)", str2lat["sin(theta)"])
d_gb = my_quant("d_gb", str2lat["d_gb"])
d_rot = my_quant("d_rot", str2lat["d_rot"])
mean_delta_bl = my_quant("mean_delta_bl", str2lat["mean_delta_bl"], r"$\AA$")
G = my_quant("G", str2lat["G"], r"J m$^{-3}$")
e_gb = my_quant("e_gb", r"$E_{GB}$", r"J m$^{-2}$")
hb = my_quant("HB", "HB", "")
bdensity = my_quant("Density", r"$Density$", r"g cm$^{-3}$")
CLTE = my_quant("CLTE", r"$CLTE$", "")
mean_bl = my_quant("mean_bl", str2lat["mean_bl"], r"$\AA$")



"""Compositional describers."""

from __future__ import annotations

import json
import os

import numpy as np
import pandas as pd
from pymatgen.core import Composition, Element, Species, Structure
from sklearn.decomposition import PCA, KernelPCA

from maml.base import BaseDescriber, describer_type
from maml.utils import Stats, get_full_stats_and_funcs, to_composition

from ._matminer import wrap_matminer_describer

CWD = os.path.abspath(os.path.dirname(__file__))

DATA_MAPPING = {
    "megnet_1": "data/elemental_embedding_1MEGNet_layer.json",
    "megnet_3": "data/elemental_embedding_3MEGNet_layer.json",
}

for length in [2, 3, 4, 8, 16, 32]:
    DATA_MAPPING[f"megnet_l{length}"] = f"data/elemental_embedding_1MEGNet_layer_length_{length}.json"
    DATA_MAPPING[f"megnet_ion_l{length}"] = f"data/ion_embedding_1MEGNet_layer_length_{length}.json"


try:
    from matminer.featurizers.composition import (  # noqa
        ElementProperty as MatminerElementProperty,
    )

    ElementProperty = wrap_matminer_describer(
        "ElementProperty", MatminerElementProperty, to_composition, describer_type="composition"
    )

except ImportError:
    ElementProperty = None


@describer_type("composition")
class ElementStats(BaseDescriber):
    """
    Element statistics. The allowed stats are accessed via ALLOWED_STATS class
    attributes. If the stats have multiple parameters, the positional arguments
    are separated by ::, e.g., moment::1::None.
    """

    ALLOWED_STATS = Stats.allowed_stats  # type: ignore
    AVAILABLE_DATA = list(DATA_MAPPING.keys())

    def __init__(
        self,
        element_properties: dict,
        stats: list[str] | None = None,
        property_names: list[str] | None = None,
        feature_batch: str = "pandas_concat",
        **kwargs,
    ):
        """
        Elemental stats for composition/str/structure.

        Args:
            element_properties (dict): element properties, e.g.,
                {'H': [0.1, 0.2, 0.3], 'He': [0.12, ...]}
            stats (list): list of stats, check ElementStats.ALLOWED_STATS
                for supported stats. The stats that support additional
                Keyword args, use ':' to separate the args. For example,
                'moment:0:None' will calculate moment stats with order=0,
                and max_order=None.
            property_names (list): list of property names, has to be consistent
                in length with properties in element_properties
            feature_batch (str): way to batch a list of feature outputs into a single
                one
            **kwargs: optional parameters include
                num_dim (int): number of dimension to keep
                reduction_algo (str): dimensional reduction algorithm
                reduction_params (dict): kwargs for dimensional reduction algorithm
        """
        num_dim = kwargs.pop("num_dim", None)
        reduction_algo = kwargs.pop("reduction_algo", "pca")
        reduction_params = kwargs.pop("reduction_params", {})

        element_properties, property_names = self._reduce_dimension(
            element_properties=element_properties,
            property_names=property_names,
            num_dim=num_dim,
            reduction_algo=reduction_algo,
            reduction_params=reduction_params,
        )

        self.element_properties = element_properties
        properties = list(self.element_properties.values())

        n_property = list({len(i) for i in properties})
        if len(n_property) > 1:
            raise ValueError("Property length not consistent")
        n_single_property = n_property[0]

        if property_names is None:
            property_names = [f"p{i}" for i in range(n_single_property)]

        if len(property_names) != n_single_property:
            raise ValueError("Property name length is not consistent")

        all_property_names = []

        if stats is None:
            stats = ["mean", "max", "min", "range", "std", "mode"]

        full_stats, stats_func = get_full_stats_and_funcs(stats)
        for stat in full_stats:
            all_property_names.extend([f"{p}_{stat}" for p in property_names])
        self.stats = full_stats
        self.element_properties = element_properties
        self.property_names = property_names
        self.n_features = len(property_names)
        self.all_property_names = all_property_names
        self.stats_func = stats_func
        super().__init__(feature_batch=feature_batch, **kwargs)

    def transform_one(self, obj: Structure | str | Composition) -> pd.DataFrame:
        """
        Transform one object, the object can be string, Compostion or Structure.

        Args:
            obj (str/Composition/Structure): object to transform

        Returns: pd.DataFrame with property names as column names

        """
        comp = to_composition(obj)
        element_n_dict = {str(i): j for i, j in comp._data.items()}
        # it is more stable when element fraction is extremely small
        # Previously, this was `comp.to_data_dict['unit_cell_composition']`

        data = []
        weights = []
        for i, j in element_n_dict.items():
            data.append(self.element_properties[i])
            weights.append(j)

        data = list(zip(*data, strict=False))
        features = []
        for stat in self.stats_func:
            for d in data:
                f = stat(d, weights)
                features.append(f)
        return pd.DataFrame([features], columns=self.all_property_names)

    @classmethod
    def from_file(cls, filename: str, stats: list[str] | None = None, **kwargs) -> ElementStats:
        """ElementStats from a json file of element property dictionary.

        The keys required are:

            element_properties
            property_names

        Args:
            filename (str): filename
            stats (list): list of stats, check ElementStats.ALLOWED_STATS
                for supported stats. The stats that support additional
                Keyword args, use ':' to separate the args. For example,
                'moment:0:None' will calculate moment stats with order=0,
                and max_order=None.
            **kwargs: Passthrough to class init.

        Returns: ElementStats class
        """
        with open(filename) as f:
            d = json.load(f)

        property_names = d.get("property_names", None)
        element_properties = d.get("element_properties") if "element_properties" in d else d
        is_element = _keys_are_elements(element_properties)
        if not is_element:
            raise ValueError("File is not in correct format")

        if "stats" in d:
            stats = d.get("stats")

        return cls(element_properties=element_properties, property_names=property_names, stats=stats, **kwargs)

    @classmethod
    def from_data(cls, data_name: list[str] | str, stats: list[str] | None = None, **kwargs) -> ElementStats:
        """
        ElementalStats from existing data file.

        Args:
            data_name (str of list of str): data name. Current supported data are
                available from ElementStats.AVAILABLE_DATA
            stats (list): list of stats, use ElementStats.ALLOWED_STATS to
                check available stats
            **kwargs: Passthrough to class init.

        Returns: ElementStats instance
        """
        if isinstance(data_name, str):
            if data_name not in ElementStats.AVAILABLE_DATA:
                raise ValueError(f"Data name not found in the list {ElementStats.AVAILABLE_DATA!s}")

            filename = os.path.join(CWD, DATA_MAPPING[data_name])
            return cls.from_file(filename, stats=stats, **kwargs)

        if isinstance(data_name, list) and len(data_name) == 1:
            return cls.from_data(data_name[0], stats=stats, **kwargs)

        property_names = []
        instances = []
        for data_name_ in data_name:
            instance = cls.from_data(data_name_, stats=stats)
            instances.append(instance)

        elements = [set(i.element_properties.keys()) for i in instances]

        common_keys = elements[0]
        for e in elements[1:]:
            common_keys.intersection_update(e)

        element_properties: dict = {i: [] for i in common_keys}

        for index, instance in enumerate(instances):
            for k in common_keys:
                element_properties[k].extend(instance.element_properties[k])

            property_names.extend([f"{index}_{i}" for i in instance.property_names])

        return cls(element_properties=element_properties, property_names=property_names, stats=stats, **kwargs)

    @staticmethod
    def _reduce_dimension(
        element_properties,
        property_names,
        num_dim: int | None = None,
        reduction_algo: str | None = "pca",
        reduction_params: dict | None = None,
    ) -> tuple[dict, list[str]]:
        """
        Reduce the feature dimension by reduction_algo.

        Args:
            element_properties (dict): dictionary of elemental/specie propeprties
            property_names (list): list of property names
            num_dim (int): number of dimension to keep
            reduction_algo (str): algorithm for dimensional reduction, currently support
                pca, kpca
            reduction_params (dict): kwargs for reduction algorithm

        Returns: new element_properties and property_names

        """
        if num_dim is None:
            return element_properties, property_names

        value_array = []
        p_keys = []
        for i, j in element_properties.items():
            value_array.append(j)
            p_keys.append(i)
        value_np_array = np.array(value_array)

        if reduction_algo == "pca":
            m = PCA(n_components=num_dim, **reduction_params)
            property_names = [f"pca_{i}" for i in range(num_dim)]
        elif reduction_algo == "kpca":
            m = KernelPCA(n_components=num_dim, **reduction_params)
            property_names = [f"kpca_{i}" for i in range(num_dim)]
        else:
            raise ValueError("Reduction algorithm not available")

        transformed_values = m.fit_transform(value_np_array)

        for key, value_list in zip(p_keys, transformed_values, strict=False):
            element_properties[key] = value_list.tolist()
        return element_properties, property_names


def _keys_are_elements(dic: dict) -> bool:
    keys = list(dic.keys())

    return all(_is_element_or_specie(key) for key in keys)


def _is_element_or_specie(s: str) -> bool:
    if s in ["D", "D+", "D-", "T"]:
        return True
    try:
        _ = Element(s)
    except ValueError:
        try:
            _ = Species.from_str(s)
        except ValueError:
            print(s)
            return False
    return True



"""Implementation of stratefied sampling approaches."""

from __future__ import annotations

import logging
import warnings

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class SelectKFromClusters(BaseEstimator, TransformerMixin):
    """Wrapper around selection of k data from each cluster."""

    def __init__(
        self,
        k: int = 1,
        allow_duplicate=False,
        selection_criteria="center",
        n_sites=None,
    ):
        """
        Args:
            k: Select k structures from each cluster.
            allow_duplicate: Whether structures are allowed to be selected over once.
            selection_criteria: The criteria to do stratified sampling from each cluster. Supported criterion
                include "random", "smallest", and "center" (default). By default, structures are ranked with respect to
                their distances to the centroid of the cluster they locate, then up to k structures with fixed ranking
                intervals are selected from each cluster, and when k=1, the structure with the smallest Euclidean
                distance to the centroid of each cluster is sampled. For "random", k structures are randomly sampled
                with replacement. For "smallest", it is ensured to select the k structures with the least number of
                atoms in each cluster.
            n_sites: The number of sites in all the structures to sample from. Only needed when
                selection_criteria="smallest".
        """
        self.k = k
        self.allow_duplicate = allow_duplicate
        allowed_selection_criterion = ["random", "smallest", "center"]
        if selection_criteria not in allowed_selection_criterion:
            raise ValueError(f"Invalid selection_criteria, it must be one of {allowed_selection_criterion}.")
        if selection_criteria == "smallest" and not n_sites:
            raise ValueError('n_sites must be provided when selection_criteria="smallest."')
        self.selection_criteria = selection_criteria
        self.n_sites = n_sites

    def fit(self, X, y=None):
        """
        Fit the model.

        Args:
            X: Input features
            y: Target.
        """
        return self

    def transform(self, clustering_data: dict):
        """
        Perform stratified sampling of data from each cluster
        based on clustering results.

        Args:
            clustering_data: Results from clustering in a dict. The dict
                should at least contain "PCAfeatures" and their respective
                "labels" of belonged clusters. The positions of centroid
                for each cluster should also be provided with "label_centers",
                with which data in each cluster can be ranked according to
                their Euclidean distance to centroid and then selected by
                interval for optimal coverage.

        Returns:
            A dict with "PCAfeatures" used in clustering and "selected_indexes"
                as the indexes of DIRECT sampled structures.
        """
        if any(key not in clustering_data for key in ["labels", "PCAfeatures"]):
            raise Exception(
                "The data returned by clustering step should at least provide label and feature information."
            )
        if self.selection_criteria == "center" and "label_centers" not in clustering_data:
            warnings.warn(
                "Centroid location is not provided, so random selection from each cluster will be performed, "
                "which likely will still outperform manual sampling in terms of feature coverage. "
            )
        if self.selection_criteria == "smallest":
            try:
                assert len(self.n_sites) == len(clustering_data["PCAfeatures"])
            except Exception:
                raise ValueError("n_sites must have same length as features processed in clustering.")

        selected_indexes = []
        for label in set(clustering_data["labels"]):
            indexes_same_label = np.where(label == clustering_data["labels"])[0]
            features_same_label = clustering_data["PCAfeatures"][indexes_same_label]
            n_same_label = len(features_same_label)
            if "label_centers" in clustering_data and self.selection_criteria == "center":
                center_same_label = clustering_data["label_centers"][label]
                distance_to_center = np.linalg.norm(features_same_label - center_same_label, axis=1).reshape(
                    len(indexes_same_label)
                )
                select_k_indexes = np.array([int(i) for i in np.linspace(0, n_same_label - 1, self.k)])
                selected_indexes.extend(
                    indexes_same_label[np.argpartition(distance_to_center, select_k_indexes)[select_k_indexes]]
                )
            elif self.selection_criteria == "smallest":
                if self.k >= n_same_label:
                    selected_indexes.extend(indexes_same_label)
                else:
                    select_k_indexes = np.arange(self.k)
                    selected_indexes.extend(
                        indexes_same_label[
                            np.argpartition(
                                np.array(self.n_sites)[indexes_same_label],
                                select_k_indexes,
                            )[select_k_indexes]
                        ]
                    )
            else:
                selected_indexes.extend(indexes_same_label[np.random.randint(n_same_label, size=self.k)])
        n_duplicate = len(selected_indexes) - len(set(selected_indexes))
        if not self.allow_duplicate and n_duplicate > 0:
            selected_indexes = list(set(selected_indexes))
        elif self.allow_duplicate and n_duplicate > 0:
            warnings.warn(f"There are {n_duplicate} duplicated selections.")
        logger.info(f"Finally selected {len(selected_indexes)} configurations.")
        return {
            "PCAfeatures": clustering_data["PCAfeatures"],
            "selected_indexes": selected_indexes,
        }
